import os
import abc
import traceback
from ._transformers_layer_group import split_into_layer_groups


HAS_TRANSFORMER = True

try:
    import torch.nn as nn
    from transformers import AutoConfig, AutoTokenizer
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False

transformer_seq_length = 512


def infer_model_type(model_name, transformer_architectures):
    model_type = "Others"
    model_name = model_name.split("/")[-1]
    for architecture in sorted(transformer_architectures, key=len, reverse=True):
        if model_name.startswith(architecture.lower()) or model_name.endswith(
            architecture.lower()
        ):
            model_type = architecture.lower()
            break
        elif model_name.startswith("opus-mt"):
            model_type = "marian"
            break
    return model_type


class ModelBackbone:
    def __init__(self, name):
        self._name = name

    @property
    def __name__(self):
        return f"{ModelBackbone.__name__}: {self._name}"


class ArcGISTransformer(nn.Module, metaclass=abc.ABCMeta):
    _outfile = "model_architecture.json"

    def __init__(
        self,
        architecture,
        pretrained_model_name,
        config=None,
        pretrained_model_path=None,
        task=None,
    ):
        super(ArcGISTransformer, self).__init__()
        self._transformer_architecture = architecture
        self._transformer_pretrained_model_name = pretrained_model_name
        self._pretrained_model_path = pretrained_model_path
        self._config = config
        self._task = task
        self._tokenizer = None
        self._transformer = None
        self._layer_groups = []

    def _load_config(self):
        """
        Load the model config from the specified path if not passed
        in the constructor and call the post processing function on it.
        """
        if self._config is None:
            if self._pretrained_model_path is None:
                raise Exception(
                    "Error, either provide model config while initializing the object or "
                    "provide pretrained-model-path containing config.json"
                )
            else:
                config_file_path = os.path.join(
                    self._pretrained_model_path, "config.json"
                )
                if os.path.exists(config_file_path):
                    self._config = AutoConfig.from_pretrained(config_file_path)
                else:
                    raise Exception(
                        f"Error, config.json not present at {config_file_path}"
                    )
        self._process_config()

    def _load_layer_groups(self):
        """
        Split the Transformer model into layer groups. Appropriate method
        will be called to split the model into different layers depending
        on the model architecture and the task selected
        """
        self._layer_groups = split_into_layer_groups(
            self._transformer, self._transformer_architecture, self._task
        )

    def get_layer_groups(self):
        """
        Returns the layer groups of the transformer architecture.
        :return: A list of list containing the layer group information of the model
        """
        return self._layer_groups

    def _load_tokenizer(self):
        """
        Loads the appropriate tokenizer for tokenizing the text
        depending on the transformer model-name parameter
        """
        # FastTokenizer not working for EntityRecognizer in transformers library version 4.5.1
        use_fast = False if self._task == "ner" else True
        self._tokenizer = AutoTokenizer.from_pretrained(
            self._transformer_pretrained_model_name,
            config=self._config,
            use_fast=use_fast,
        )
        self._set_max_seq_length()

    def init_model(self):
        """
        A wrapper function to load necessary attributes like model-config,
        model-tokeniser, transformer-model and layer-groups
        """
        self._load_config()
        self._load_tokenizer()
        self._load_transformer()
        self._load_layer_groups()

    @abc.abstractmethod
    def _process_config(self):
        """
        Post processing method applied to the model config. Class inheriting this
        class will apply their specific post processing on model config
        """
        pass

    @abc.abstractmethod
    def _set_max_seq_length(self):
        """
        Set the max sequence length parameter. This is useful in cases where a transformer
        transformer backbone doesn't have a max sequence length parameter for example XLNet,
        Longformer, Bart etc. This can leads to out of memory issue while training on GPUs
        """
        pass

    @abc.abstractmethod
    def _load_transformer(self):
        """
        Method to load the appropriate transformer model. Class inheriting this
        class will load specific model architectures to solve specific nlp tasks
        """
        pass

    @classmethod
    def load_pretrained_model(cls, path):
        """
        Method to load the fine-tuned model which was saved on the disk

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The disk location from
                                where the fine-tuned model has to be loaded
        ---------------------   -------------------------------------------
        """
        pass

    @abc.abstractmethod
    def save(self, model_path):
        """
        Method to save the fine-tuned model to the disk

        =====================   =================================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------------
        model_path              Required string. The disk location where the
                                fine-tuned model has to be saved
        ---------------------   -------------------------------------------------
        """
        pass

    @abc.abstractmethod
    def forward(self, input_ids, **kwargs):
        pass



# Variation of TFLite Model Maker Object Detector
# https://github.com/tensorflow/examples/tree/master/tensorflow_examples/lite/model_maker
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from email.mime import image
import math
from dataclasses import dataclass
from re import S
from typing import Optional, Union


from ._arcgis_model import SaveModelCallback
from .._utils.env import HAS_TENSORFLOW, ARCGIS_ENABLE_TF_BACKEND

try:
    import os
    from types import MethodType
    import numpy as np
    from fastai.basics import *
    import tensorflow as tf

    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
    import absl.logging

    absl.logging.set_verbosity(absl.logging.ERROR)
    from fastai.vision.data import ImageDataBunch
    from fastai.callbacks import EarlyStoppingCallback, OneCycleScheduler
    from fastai.callback import annealing_cos
    from fastai.core import ifnone
    import tensorflow as tf
    from tensorflow_examples.lite.model_maker.core.task.object_detector import (
        ObjectDetector,
    )
    from tensorflow_examples.lite.model_maker.core.task.configs import (
        QuantizationConfig,
    )
    from tensorflow_examples.lite.model_maker.core.task.custom_model import _get_params
    from tensorflow_examples.lite.model_maker.core.task import model_spec
    from tensorflow_examples.lite.model_maker.core import compat
    from tensorflow_examples.lite.model_maker.core.data_util import (
        object_detector_dataloader,
        dataloader,
    )
    from tensorflow_examples.lite.model_maker.core.task import model_spec as ms
    from tensorflow_examples.lite.model_maker.core.task.model_spec import (
        object_detector_spec,
    )
    from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import (
        train,
        util_keras,
    )
    from tensorflow_examples.lite.model_maker.core.data_util.object_detector_dataloader import (
        DataLoader as tflite_data_loader,
    )
    from lxml import etree
    from tensorflow_examples.lite.model_maker.third_party.efficientdet.dataset import (
        tfrecord_util,
    )

    HAS_FASTAI = True
except:
    HAS_FASTAI = False


def check_data_sanity(data, dataset_types=[]):
    try:
        if isinstance(data, ImageDataBunch):
            for k, v in data.class_mapping.items():
                if k == 0 and v != "background":
                    import warnings

                    warnings.warn("class 0 is reserved only for background\n")
                    return False
            return True
        else:
            return False
    except Exception as e:
        return False


def _get_image_tensor(image):
    return tf.expand_dims(tf.convert_to_tensor(np.array(image)), axis=0)


def _get_ann_files(data):
    if data._is_empty:
        return [None, None]

    val_split_pct = data._val_split_pct
    labels_path = os.path.join(data.orig_path, "labels")
    if not os.path.isdir(labels_path):
        return [None, None]
    else:
        tot_files = len(os.listdir(labels_path))
        files = sorted(os.listdir(labels_path))
        files = [f[:-4] for f in files if os.path.isfile(os.path.join(labels_path, f))]
        num_val_files = min(
            max(math.floor((val_split_pct * tot_files)), 1), tot_files - 1
        )
        train_files = files[:-num_val_files]
        val_files = files[-num_val_files:]

        return [train_files, val_files]


def _calculate_avg_loss(avg_loss, loss, beta=0.98):
    return beta * avg_loss + (1 - beta) * loss


def _calculate_smoothed_loss(avg_loss, iteration, beta=0.98):
    return avg_loss / (1 - beta**iteration)


class EmptyLoader(dataloader.DataLoader):
    def __init__(self, label_map):
        super(EmptyLoader, self).__init__(dataset=None, size=1)
        self.label_map = label_map


def _get_tf_data_loader(data, ann_files=None):
    dataset_type = None
    if hasattr(data, "dataset_type"):
        dataset_type = getattr(data, "dataset_type")
        if data._is_empty:
            return EmptyLoader(label_map=data.class_mapping)

        if dataset_type == "PASCAL_VOC_rectangles":
            try:
                return tflite_data_loader.from_pascal_voc(
                    os.path.join(data.orig_path, "images"),
                    os.path.join(data.orig_path, "labels"),
                    label_map=data.class_mapping,
                    annotation_filenames=ann_files,
                )
            except:
                mapping = dict()
                for k in data.class_mapping:
                    mapping[k] = str(k)
                return tflite_data_loader.from_pascal_voc(
                    os.path.join(data.orig_path, "images"),
                    os.path.join(data.orig_path, "labels"),
                    label_map=mapping,
                    annotation_filenames=ann_files,
                )
        elif dataset_type == "KITTI_rectangles":
            return None
        else:
            dataset_type = None
    if dataset_type == None:
        raise Exception(f"Enter only compatible datasets\n")


class ConstantLrSchedule(tf.optimizers.schedules.LearningRateSchedule):
    """Constant learning rate schedule."""

    def __init__(self, lr):
        """Build a ConstantLrSchedule."""
        super().__init__()
        self.default_lr = defaults.lr
        self.lr = lr

    def restart(self):
        self.lr = defaults.lr

    def __call__(self, step):
        return self.lr


class AnnealedLrSchedule(tf.optimizers.schedules.LearningRateSchedule):
    """Annealed learning rate schedule."""

    def __init__(self, vals: Union[tuple, int], n_iter: int):
        """Build a AnnealedLrSchedule."""
        super().__init__()
        self.start, self.end = (
            (vals[0], vals[1]) if isinstance(vals, tuple) else (vals, 0)
        )
        self.n_iter = max(1, n_iter)
        self.func = AnnealedLrSchedule.annealing_exp
        self.n = 0
        self.lr = self.start

    @classmethod
    def annealing_exp(cls, start: float, end: float, pct: float) -> float:
        "Exponentially anneal from `start` to `end` as pct goes from 0.0 to 1.0."
        return start * (end / start) ** pct

    def restart(self):
        self.n = 0

    def step(self):
        "Return next value along annealed schedule."
        self.n += 1
        self.lr = AnnealedLrSchedule.annealing_exp(
            self.start, self.end, self.n / self.n_iter
        )

    @property
    def is_done(self) -> bool:
        "Return `True` if schedule completed."
        return self.n >= self.n_iter

    def __call__(self, step):
        return self.lr


from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils


def _get_optimizer(params):
    """Get optimizer."""
    try:
        learning_rate = train_lib.learning_rate_schedule(params)
    except:
        if params["lr_decay_method"] == "annealed":
            learning_rate = AnnealedLrSchedule(
                (params["start_lr"], params["end_lr"]), params["num_it"]
            )
            learning_rate = params["start_lr"]
        else:
            learning_rate = params["constant_lr"]

    momentum = params["momentum"]
    if params["optimizer"].lower() == "sgd":
        optimizer = tf.keras.optimizers.SGD(learning_rate, momentum=momentum)
    elif params["optimizer"].lower() == "adam":
        optimizer = tf.keras.optimizers.Adam(learning_rate, beta_1=momentum)
    else:
        raise ValueError("optimizers should be adam or sgd")

    moving_average_decay = params["moving_average_decay"]
    if moving_average_decay:
        from tensorflow_addons import optimizers as tfa_optimizers

        optimizer = tfa_optimizers.MovingAverage(
            optimizer, average_decay=moving_average_decay, dynamic_decay=True
        )
    precision = utils.get_precision(params["strategy"], params["mixed_precision"])
    if precision == "mixed_float16" and params["loss_scale"]:
        optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
            optimizer, initial_scale=params["loss_scale"]
        )
    return optimizer


def _setup_model(model, config, build=True):
    """Build and compile model."""
    if build is True:
        model.build((None, *config.image_size, 3))
    model.compile(
        steps_per_execution=config.steps_per_execution,
        optimizer=_get_optimizer(config.as_dict()),
        loss={
            train_lib.BoxLoss.__name__: train_lib.BoxLoss(
                config.delta, reduction=tf.keras.losses.Reduction.NONE
            ),
            train_lib.BoxIouLoss.__name__: train_lib.BoxIouLoss(
                config.iou_loss_type,
                config.min_level,
                config.max_level,
                config.num_scales,
                config.aspect_ratios,
                config.anchor_scale,
                config.image_size,
                reduction=tf.keras.losses.Reduction.NONE,
            ),
            train_lib.FocalLoss.__name__: train_lib.FocalLoss(
                config.alpha,
                config.gamma,
                label_smoothing=config.label_smoothing,
                reduction=tf.keras.losses.Reduction.NONE,
            ),
            tf.keras.losses.SparseCategoricalCrossentropy.__name__: tf.keras.losses.SparseCategoricalCrossentropy(
                from_logits=True, reduction=tf.keras.losses.Reduction.NONE
            ),
        },
    )
    return model


class EfficientDetTrainer(ObjectDetector):
    def __init__(
        self,
        model_spec: object_detector_spec.EfficientDetModelSpec,
        label_map: Dict[int, str],
        representative_data: Optional[object_detector_dataloader.DataLoader] = None,
    ) -> None:
        self._data = representative_data
        self._config = None
        self._ann_files = _get_ann_files(self._data)
        self._data_loader = _get_tf_data_loader(representative_data, self._ann_files[0])
        self._val_loader = _get_tf_data_loader(representative_data, self._ann_files[1])
        super().__init__(model_spec, label_map, self._data_loader)
        self._build_model = True

    def setup_model(self):
        self._config = self.model_spec.config
        batch_size = self._data.batch_size
        self._train_ds = None
        self._valid_ds = None
        self._steps_per_epoch = None

        self._valid_steps_per_epoch = None
        self._validation_steps = None
        with self.model_spec.ds_strategy.scope():
            if not self._data._is_empty:
                self._train_ds, self._steps_per_epoch, _ = self._get_dataset_and_steps(
                    self._data_loader, batch_size, is_training=True
                )
                (
                    self._valid_ds,
                    self._validation_steps,
                    val_json_file,
                ) = self._get_dataset_and_steps(
                    self._val_loader, batch_size, is_training=False
                )
                self._config.update(
                    dict(
                        steps_per_epoch=self._steps_per_epoch,
                        eval_samples=batch_size * self._validation_steps,
                        batch_size=batch_size,
                        val_json_file=val_json_file,
                    )
                )  # validation
            else:
                self._config.update(
                    dict(batch_size=batch_size, steps_per_epoch=batch_size)
                )
            _setup_model(self.model, self._config, self._build_model)
            self._build_model = False
            train.init_experimental(self._config)

    def get_opt_func(self):
        self._config = self.model_spec.config.as_dict()
        if self._config["optimizer"].lower() == "sgd":
            optimizer = tf.keras.optimizers.SGD
        elif self._config["optimizer"].lower() == "adam":
            optimizer = tf.keras.optimizers.Adam
        else:
            raise ValueError("optimizers should be adam or sgd")

        return optimizer

    def get_tf_dataset(self, data_loader, batch_size=16):
        with self.model_spec.ds_strategy.scope():
            ds, steps_per_epoch, _ = self._get_dataset_and_steps(
                data_loader, batch_size, is_training=True
            )
            return ds

    def fit(
        self,
        epochs: int,
        lr: Union[Floats, slice] = defaults.lr,
        wd: Floats = None,
        callbacks: Collection[Callback] = None,
        annealed_schedule=False,
        start_lr: float = 1e-7,
        end_lr: float = 10,
        num_it: int = 100,
        freeze_model: bool = True,
        one_cycle: bool = False,
    ) -> None:
        if self._data._is_empty:
            print("\n\nInvalid Fit\n\n")
            return

        self._lr_decay_method = self.model_spec.config.lr_decay_method
        if annealed_schedule is True:
            self.model_spec.config.lr_decay_method = "annealed"
            self.model_spec.config.start_lr = start_lr
            self.model_spec.config.end_lr = end_lr
            self.model_spec.config.num_it = num_it
        elif one_cycle:
            self.model_spec.config.lr_decay_method = "constant"
            self.model_spec.config.constant_lr = lr

        if freeze_model is False:
            self.model.config.var_freeze_expr = None
            self.model_spec.config.var_freeze_expr = None
        else:
            self.model.config.var_freeze_expr = "(efficientnet|fpn_cells|resample_p6)"
            self.model_spec.config.var_freeze_expr = (
                "(efficientnet|fpn_cells|resample_p6)"
            )

        num_epochs = self.model_spec.config.num_epochs
        learning_rate = self.model_spec.config.learning_rate
        self.model_spec.config.num_epochs = epochs
        self.model_spec.config.learning_rate = lr

        self.setup_model()

        with self.model_spec.ds_strategy.scope():
            self.model.fit(
                self._train_ds,
                epochs=epochs,
                steps_per_epoch=self._steps_per_epoch,
                verbose=0,
                callbacks=callbacks,
                validation_data=self._valid_ds,
                validation_steps=self._validation_steps,
            )
        self.model_spec.config.lr_decay_method = self._lr_decay_method
        self.model_spec.config.num_epochs = num_epochs
        self.model_spec.config.learning_rate = learning_rate

    def _export_saved_model(self, saved_model_dir: str) -> None:
        model = self.model
        original_optimizer = model.optimizer
        model.optimizer = None

        tf.saved_model.save(model, saved_model_dir)

        model.optimizer = original_optimizer

    def _save_weights(
        self, name, path, model_dir, quantized=False, tflite=True, **kwargs
    ):
        import os

        if self._data._is_empty:
            quantized = False

        model_save_path = path / model_dir / f"{name}.tflite"
        tflite_filename = f"{name}.tflite"
        saved_model_dir = f"weights"
        export_dir = path / model_dir
        if not tf.io.gfile.exists(export_dir):
            tf.io.gfile.makedirs(export_dir)

        if quantized is False:
            kwargs["quantized_config"] = None
            if self._data._is_empty:
                kwargs["quantization_config"] = None
        else:
            kwargs["quantized_config"] = "default"
            kwargs["quantization_config"] = "default"

        tflite_filepath = os.path.join(export_dir, tflite_filename)
        export_tflite_kwargs, kwargs = _get_params(self._export_tflite, **kwargs)

        if tflite:
            self._export_tflite(tflite_filepath, **export_tflite_kwargs)
        else:
            model_save_path = path / model_dir / f"{name}.pb"

        saved_model_dir = os.path.join(export_dir, saved_model_dir)
        export_saved_model_kwargs, kwargs = _get_params(
            self._export_saved_model, **kwargs
        )
        self._export_saved_model(saved_model_dir, **export_saved_model_kwargs)
        if not tflite:
            import shutil

            shutil.copyfile(
                os.path.join(saved_model_dir, "saved_model.pb"),
                os.path.join(export_dir, f"{name}.pb"),
            )

        return model_save_path

    def compute_metrics(self):
        metrics = self.evaluate(self._val_loader)
        metrics.update(
            (key, value * 1.0) if value >= 0.0 else (key, np.float64(0.0))
            for key, value in metrics.items()
        )
        return metrics

    def _get_processed_output(self, output, thresh=0.4):
        boxes = output[0]
        scores = output[1]
        labels = output[2]

        detections = [
            boxes[:, :, 0],
            boxes[:, :, 1],
            boxes[:, :, 2],
            boxes[:, :, 3],
            scores,
            labels,
        ]

        detections = tf.stack(detections, axis=-1)[0].numpy()
        boxes = detections[:, 0:4]
        classes = detections[:, 5].astype(int).tolist()
        scores_ = detections[:, 4].tolist()

        predictions = []
        labels = []
        scores = []

        for i in range(boxes.shape[0]):
            if scores_ is None or scores_[i] > thresh:
                ymin, xmin, ymax, xmax = boxes[i].tolist()
                box = [xmin, ymin, xmax - xmin, ymax - ymin]
                predictions.append(box)
                labels.append(classes[i])
                scores.append(scores_[i])

        return predictions, labels, scores

    def predict(self, image, thresh, nms_overlap):
        original_optimizer = self.model.optimizer
        original_thresh = self.model_spec.config.nms_configs.iou_thresh
        self.model.optimizer = None
        self.model_spec.config.nms_configs.iou_thresh = thresh
        export_model = object_detector_spec.ExportModel(
            self.model, self.model_spec.config
        )
        self.model.optimizer = original_optimizer
        self.model_spec.config.nms_configs.iou_thresh = original_thresh
        output = export_model(_get_image_tensor(image))
        return self._get_processed_output(output, thresh)

    @classmethod
    def create(cls, train_data, backbone_name: str, train_whole_model: bool = False):
        """Creates a model for object detection.

        Args:
            train_data: Training data.
            backbone_name: BackBone name for the model.
            batch_size: Batch size for training.
            train_whole_model: Boolean, False by default. If true, train the whole
            model. Otherwise, only train the layers that are not match
            `model_spec.config.var_freeze_expr`.

        Returns:
            An instance based on ObjectDetector.
        """
        spec = model_spec.get(backbone_name)
        spec = ms.get(spec)
        spec.config.batch_size = train_data.batch_size
        if train_whole_model:
            spec.config.var_freeze_expr = None
        if compat.get_tf_behavior() not in spec.compat_tf_versions:
            raise ValueError(
                "Incompatible versions. Expect {}, but got {}.".format(
                    spec.compat_tf_versions, compat.get_tf_behavior()
                )
            )

        object_detector = cls(spec, train_data.class_mapping, train_data)
        with object_detector.model_spec.ds_strategy.scope():
            object_detector.create_model()
        object_detector.model.train_step = MethodType(train_step, object_detector.model)
        object_detector.setup_model()
        return object_detector


def train_step(self, data):
    """Train step.

    Args:
        data: Tuple of (images, labels). Image tensor with shape [batch_size,
        height, width, 3]. The height and width are fixed and equal.Input labels
        in a dictionary. The labels include class targets and box targets which
        are dense label maps. The labels are generated from get_input_fn
        function in data/dataloader.py.

    Returns:
        A dict record loss info.
    """
    images, labels = data
    if self.config.img_summary_steps:
        with self.summary_writer.as_default():
            tf.summary.image("input_image", images)
    with tf.GradientTape() as tape:
        if len(self.config.heads) == 2:
            cls_outputs, box_outputs, seg_outputs = util_keras.fp16_to_fp32_nested(
                self(images, training=True)
            )
            loss_dtype = cls_outputs[0].dtype
        elif "object_detection" in self.config.heads:
            cls_outputs, box_outputs = util_keras.fp16_to_fp32_nested(
                self(images, training=True)
            )
            loss_dtype = cls_outputs[0].dtype
        elif "segmentation" in self.config.heads:
            (seg_outputs,) = util_keras.fp16_to_fp32_nested(self(images, training=True))
            loss_dtype = seg_outputs.dtype
        else:
            raise ValueError("No valid head found: {}".format(self.config.heads))
        labels = util_keras.fp16_to_fp32_nested(labels)

        total_loss = 0
        loss_vals = {}
        if "object_detection" in self.config.heads:
            det_loss = self._detection_loss(cls_outputs, box_outputs, labels, loss_vals)
            total_loss += det_loss
        if "segmentation" in self.config.heads:
            seg_loss_layer = self.loss[
                tf.keras.losses.SparseCategoricalCrossentropy.__name__
            ]
            seg_loss = seg_loss_layer(labels["image_masks"], seg_outputs)
            total_loss += seg_loss
            loss_vals["seg_loss"] = seg_loss

        reg_l2_loss = self._reg_l2_loss(self.config.weight_decay)
        loss_vals["reg_l2_loss"] = reg_l2_loss
        total_loss += tf.cast(reg_l2_loss, loss_dtype)
        if isinstance(self.optimizer, tf.keras.mixed_precision.LossScaleOptimizer):
            scaled_loss = self.optimizer.get_scaled_loss(total_loss)
            optimizer = self.optimizer.inner_optimizer
        else:
            scaled_loss = total_loss
            optimizer = self.optimizer
    loss_vals["loss"] = total_loss
    if isinstance(
        optimizer.learning_rate, tf.optimizers.schedules.LearningRateSchedule
    ):
        loss_vals["learning_rate"] = optimizer.learning_rate(optimizer.iterations)
    else:
        loss_vals["learning_rate"] = optimizer.learning_rate

    trainable_vars = self._freeze_vars()
    scaled_gradients = tape.gradient(scaled_loss, trainable_vars)
    if isinstance(self.optimizer, tf.keras.mixed_precision.LossScaleOptimizer):
        gradients = self.optimizer.get_unscaled_gradients(scaled_gradients)
    else:
        gradients = scaled_gradients
    if self.config.clip_gradients_norm > 0:
        clip_norm = abs(self.config.clip_gradients_norm)
        gradients = [
            tf.clip_by_norm(g, clip_norm) if g is not None else None for g in gradients
        ]
        gradients, _ = tf.clip_by_global_norm(gradients, clip_norm)
        loss_vals["gradient_norm"] = tf.linalg.global_norm(gradients)
    self.optimizer.apply_gradients(zip(gradients, trainable_vars))
    return loss_vals


from .._utils.fastai_tf_fit import (
    TfLearner,
    TfRegularizer,
    tf_flatten_model,
    _pytorch_to_tf_batch,
    _pytorch_to_tf,
)
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import (
    train_lib,
)


@dataclass
class EfficientDetLearner(TfLearner):
    _trainer: EfficientDetTrainer = None
    tf_dataset: tf.data.Dataset = None
    _compute_mean_avp = False

    def __post_init__(self) -> None:
        "Setup path,metrics, callbacks and ensure model directory exists."
        if self.tf_dataset is None or self._trainer is None:
            raise ValueError("\nInvalid arguments\n")

        self.path = Path(ifnone(self.path, self.data.path))
        (self.path / self.model_dir).mkdir(parents=True, exist_ok=True)
        self.metrics = listify(self.metrics)
        self._freeze_model = True
        self.recorder = TFRecorder(0, 0)
        if not self.layer_groups:
            self.layer_groups = tf_flatten_model(self.model)

    def init(self, init):
        raise NotImplementedError

    def _save_tflite(self, name, post_processed=True, quantized=False, **kwargs):
        return self._trainer._save_weights(name, self.path, self.model_dir, quantized)

    def compute_metrics(self):
        return self._trainer.compute_metrics()

    def save(self, name, return_path=True, **kwargs):
        "Save model with `name` to `self.model_dir`."
        return self._trainer._save_weights(
            name, self.path, self.model_dir, tflite=False
        )

    def load(self, file, device=None, **kwargs):
        weights_save_path = str(
            self.path / self.model_dir / f"weights/variables/variables"
        )
        self._trainer.model.load_weights(weights_save_path)

    def fit(
        self,
        epochs: int,
        lr: Union[Floats, slice] = defaults.lr,
        wd: Floats = None,
        callbacks: Collection[Callback] = None,
        annealed_schedule=False,
        start_lr: float = 1e-7,
        end_lr: float = 10,
        num_it: int = 100,
        one_cycle: bool = False,
    ) -> None:
        lr = self.lr_range(lr)
        for i in range(0, len(callbacks)):
            if isinstance(callbacks[i], EarlyStoppingCallback):
                callbacks[i] = tf.keras.callbacks.EarlyStopping(
                    min_delta=0.001, patience=5
                )
            if isinstance(callbacks[i], SaveModelCallback):
                callbacks[i] = CheckPointCallback(callback=callbacks[i])
            try:
                from .._utils.tensorboard_utils import ArcGISTBCallback

                if isinstance(callbacks[i], ArcGISTBCallback):
                    log_dir = callbacks[i]._base_dir
                    callbacks[i] = tf.keras.callbacks.TensorBoard(log_dir=log_dir)
            except:
                pass

        self.recorder = TFRecorder(epochs, len(self.data.train_dl))
        callbacks = [self.recorder] + callbacks
        self._trainer.fit(
            epochs,
            lr,
            wd,
            callbacks,
            annealed_schedule,
            start_lr,
            end_lr,
            num_it,
            freeze_model=self._freeze_model,
            one_cycle=one_cycle,
        )

    def unfreeze(self):
        self._freeze_model = False

    def freeze(self) -> None:
        "Freeze up to last layer."
        self._freeze_model = True

    def predict(self, image, thresh, nms_overlap):
        return self._trainer.predict(image, thresh, nms_overlap)

    def _get_annotations_from_xml(self, data):
        predictions = []
        if "object" in data:
            for obj in data["object"]:
                if not "difficult" in obj or obj["difficult"] == "Unspecified":
                    difficult = False
                else:
                    difficult = bool(int(obj["difficult"]))
                if difficult:
                    continue

                xmin = int(math.floor(float(obj["bndbox"]["xmin"])))
                ymin = int(math.floor(float(obj["bndbox"]["ymin"])))
                xmax = int(math.floor(float(obj["bndbox"]["xmax"])))
                ymax = int(math.floor(float(obj["bndbox"]["ymax"])))

                predictions.append([xmin, ymin, xmax - xmin, ymax - ymin])
        return predictions

    def get_gt_batches(self, nbatches, type_data_loader="validation"):
        batches = []
        data = self._trainer._data
        labels_path = os.path.join(data.orig_path, "labels")
        images_path = os.path.join(data.orig_path, "images")
        if not os.path.isdir(labels_path):
            return batches
        ann_files = self._trainer._ann_files[0]
        if type_data_loader == "validation":
            ann_files = self._trainer._ann_files[1]

        i = 0
        while i < nbatches and i < len(ann_files):
            filename = os.path.join(labels_path, ann_files[i] + ".xml")
            if os.path.isfile(filename):
                with tf.io.gfile.GFile(filename, "r") as fid:
                    xml_str = fid.read()
                xml = etree.fromstring(xml_str)
                xml_dict = tfrecord_util.recursive_parse_xml_to_dict(xml)["annotation"]
                ann = self._get_annotations_from_xml(xml_dict)
                batches.append([os.path.join(images_path, xml_dict["filename"]), ann])
                i += 1
            else:
                continue

        return batches


from time import time
from fastprogress.fastprogress import master_bar, progress_bar


def get_monitor_value(self):
    "Pick the monitored value."
    if self.monitor == "trn_loss" and len(self.learn.recorder.losses) == 0:
        return None
    elif len(self.learn.recorder.val_losses) == 0:
        return None
    values = {
        "train_loss": self.learn.recorder.losses[-1],
        "valid_loss": self.learn.recorder.val_losses[-1],
    }
    if values["valid_loss"] is None:
        return
    from warnings import warn

    if values.get(self.monitor) is None:
        warn(
            f'{self.__class__} conditioned on metric `{self.monitor}` which is not available. Available metrics are: {", ".join(map(str, self.learn.recorder.names[1:-1]))}'
        )
    return values.get(self.monitor)


class CheckPointCallback(tf.keras.callbacks.Callback):
    def __init__(self, callback):
        self._callback = callback
        self._callback.get_monitor_value = MethodType(get_monitor_value, self._callback)

    def on_train_begin(self, logs=None):
        self._callback.on_train_begin()

    def on_epoch_end(self, epoch, logs=None):
        self._callback.on_epoch_end(epoch, stop_training=self.model.stop_training)


class TFRecorder(tf.keras.callbacks.Callback):
    def __init__(self, epochs, num_batches):
        super(TFRecorder, self).__init__()
        self._iteration = 0
        self.epochs = epochs
        self.batches_per_epoch = num_batches
        self.silent = False
        self.add_time = True
        self.pbar_iter = None
        self.mbar_iter = None

    def on_train_begin(self, logs=None):
        "Initialize optimizer and learner hyperparameters."
        self.losses = []
        self.epoch_losses = []
        self.val_losses = []
        self.lrs = []
        self.nb_batches = []
        self.avg_loss = 0.0
        self.smoothed_loss = 0.0
        self.num_batch = 0
        self.pbar = master_bar(range(self.epochs))
        self.pbar_iter = iter(self.pbar)
        self.names = ["epoch", "train_loss", "valid_loss", "time"]
        self.pbar.write(self.names, table=True)

    def on_epoch_begin(self, epoch, logs=None):
        next(self.pbar_iter)
        self.start_epoch = time()
        self.mbar = progress_bar(range(self.batches_per_epoch), parent=self.pbar)
        self.mbar_iter = iter(self.mbar)

    def on_train_batch_begin(self, batch, logs=None):
        lr = 0.0
        next(self.mbar_iter)
        if isinstance(
            self.model.optimizer.learning_rate,
            tf.optimizers.schedules.LearningRateSchedule,
        ):
            lr = (float)(
                tf.keras.backend.get_value(
                    self.model.optimizer.lr(self.model.optimizer.iterations)
                )
            )
        else:
            lr = (float)(tf.keras.backend.get_value(self.model.optimizer.lr))
        self.lrs.append(lr)

    def on_train_batch_end(self, batch, logs=None):
        "Determine if loss has runaway and we should stop."
        loss = logs.get("loss")
        self.avg_loss = _calculate_avg_loss(self.avg_loss, loss)
        self.smoothed_loss = _calculate_smoothed_loss(
            self.avg_loss, self._iteration + 1
        )
        self.losses.append(self.smoothed_loss)
        self._iteration = self._iteration + 1
        self.num_batch = batch + 1
        if self.pbar is not None and hasattr(self.pbar, "child"):
            self.pbar.child.comment = f"{self.smoothed_loss:.4f}"

    def on_epoch_end(self, epoch, logs=None):
        "Record at end of epoch."
        self.val_losses.append(logs.get("val_loss"))
        self.epoch_losses.append(self.smoothed_loss)
        self.nb_batches.append(self.num_batch)
        self.format_stats([epoch, self.smoothed_loss, logs.get("val_loss")])

        try:
            next(self.mbar_iter)
        except:
            pass

    def on_train_end(self, logs=None):
        try:
            next(self.pbar_iter)
        except:
            pass


class EfficientLRFinder(tf.keras.callbacks.Callback):
    "Causes `learn` to go on a mock training from `start_lr` to `end_lr` for `num_it` iterations."

    def __init__(
        self,
        learn: Learner,
        start_lr: float = 1e-7,
        end_lr: float = 10,
        num_it: int = 100,
        stop_div: bool = True,
    ):
        super(EfficientLRFinder, self).__init__()
        self.learn = learn
        self.data, self.stop_div = learn.data, stop_div
        # To avoid validating if the train_dl has less than num_it batches, we put aside the valid_dl and remove it
        # during the call to fit.
        import copy

        self.sched = AnnealedLrSchedule((start_lr, end_lr), num_it)
        self.valid_dl = copy.deepcopy(learn.data.valid_dl)
        self.data.valid_dl = None
        self._iteration = 0

    def on_train_begin(self, logs=None):
        "Initialize optimizer and learner hyperparameters."
        self.model.save_weights("tmp")
        self.stop, self.best_loss, self.avg_loss = False, np.Inf, 0.0
        tf.keras.backend.set_value(self.model.optimizer.lr, self.sched(0))

    def on_train_batch_end(self, batch, logs=None):
        "Determine if loss has runaway and we should stop."
        loss = logs.get("loss")
        self.avg_loss = _calculate_avg_loss(self.avg_loss, loss)
        smoothed_loss = _calculate_smoothed_loss(self.avg_loss, self._iteration + 1)
        if self._iteration == 0 or smoothed_loss < self.best_loss:
            self.best_loss = smoothed_loss

        self.sched.step()
        tf.keras.backend.set_value(
            self.model.optimizer.lr, tf.keras.backend.get_value(self.sched(0))
        )

        if self.sched.is_done or (
            self.stop_div
            and (smoothed_loss > 4 * self.best_loss or np.isnan(smoothed_loss))
        ):
            self.stop = self._iteration
            # We use the smoothed loss to decide on the stopping since it's less shaky.
            self.model.stop_training = True

        self._iteration += 1

    def on_train_end(self, logs=None) -> None:
        "Cleanup learn model weights disturbed during LRFind exploration."
        # restore the valid_dl we turned off on `__init__`
        self.data.valid_dl = self.valid_dl
        self.model.load_weights("tmp")
        self._iteration = 0
        if hasattr(self.learn.model, "reset"):
            self.learn.model.reset()
        print(
            "LR Finder is complete, type {learner_name}.recorder.plot() to see the graph."
        )

    def on_epoch_end(self, epoch, logs=None):
        if self.stop:
            self.model.stop_training = True


def efficient_lr_find(
    learn: Learner,
    start_lr: Floats = 1e-7,
    end_lr: Floats = 10,
    num_it: int = 100,
    stop_div: bool = True,
    **kwargs: Any,
):
    "Explore lr from `start_lr` to `end_lr` over `num_it` iterations in `learn`. If `stop_div`, stops when loss diverges."
    start_lr = learn.lr_range(start_lr)
    start_lr = np.array(start_lr) if is_listy(start_lr) else start_lr
    end_lr = learn.lr_range(end_lr)
    end_lr = np.array(end_lr) if is_listy(end_lr) else end_lr
    cb = EfficientLRFinder(learn, start_lr, end_lr, num_it, stop_div)
    a = int(np.ceil(num_it / len(learn.data.train_dl)))
    learn.fit(
        a,
        start_lr,
        callbacks=[cb],
        annealed_schedule=True,
        end_lr=end_lr,
        num_it=num_it,
        **kwargs,
    )


class TFOneCycleScheduler(tf.keras.callbacks.Callback):
    "Manage 1-Cycle style training as outlined in Leslie Smith's [paper](https://arxiv.org/pdf/1803.09820.pdf)."

    def __init__(self, learn: EfficientDetLearner, initial_epoch=0, scheduler=None):
        super(TFOneCycleScheduler, self).__init__()
        self.learn = learn
        self._scheduler = scheduler
        self._initial_epoch = 0

    def jump_to_epoch(self, epoch: int):
        for _ in range(len(self.learn.data.train_dl) * epoch):
            self.on_train_batch_end(epoch)

    def on_train_begin(self, logs=None):
        "Initialize our optimization params based on our annealing schedule."
        epoch = self._initial_epoch
        self._scheduler.start_epoch = ifnone(self._scheduler.start_epoch, epoch)
        self._scheduler.tot_epochs = ifnone(
            self._scheduler.tot_epochs, self.params["epochs"]
        )
        n = len(self.learn.data.train_dl) * self._scheduler.tot_epochs
        a1 = int(n * self._scheduler.pct_start)
        a2 = n - a1
        self._scheduler.phases = ((a1, annealing_cos), (a2, annealing_cos))
        low_lr = self._scheduler.lr_max / self._scheduler.div_factor
        self._scheduler.lr_scheds = self._scheduler.steps(
            (low_lr, self._scheduler.lr_max),
            (
                self._scheduler.lr_max,
                self._scheduler.lr_max / self._scheduler.final_div,
            ),
        )
        self._scheduler.mom_scheds = self._scheduler.steps(
            self._scheduler.moms, (self._scheduler.moms[1], self._scheduler.moms[0])
        )
        tf.keras.backend.set_value(
            self.model.optimizer.lr,
            tf.keras.backend.get_value(self._scheduler.lr_scheds[0].start),
        )
        tf.keras.backend.set_value(
            self.model.optimizer.momentum,
            tf.keras.backend.get_value(self._scheduler.mom_scheds[0].start),
        )
        self._scheduler.idx_s = 0

        self.jump_to_epoch(epoch)

    def on_train_batch_end(self, batch, logs=None):
        "Take one step forward on the annealing schedule for the optim params."
        if self._scheduler.idx_s >= len(self._scheduler.lr_scheds):
            self.model.stop_training = True
            return
        tf.keras.backend.set_value(
            self.model.optimizer.lr,
            tf.keras.backend.get_value(
                self._scheduler.lr_scheds[self._scheduler.idx_s].step()
            ),
        )
        tf.keras.backend.set_value(
            self.model.optimizer.momentum,
            tf.keras.backend.get_value(
                self._scheduler.mom_scheds[self._scheduler.idx_s].step()
            ),
        )

        # when the current schedule is complete we move onto the next
        # schedule. (in 1-cycle there are two schedules)
        if self._scheduler.lr_scheds[self._scheduler.idx_s].is_done:
            self._scheduler.idx_s += 1

    def on_epoch_end(self, epoch, logs=None):
        "Tell Learner to stop if the cycle is finished."
        if epoch > self._scheduler.tot_epochs:
            self.model.stop_training = True


def tf_fit_one_cycle(
    learn: Learner,
    cyc_len: int,
    max_lr: Union[Floats, slice] = defaults.lr,
    moms: Tuple[float, float] = (0.95, 0.85),
    div_factor: float = 25.0,
    pct_start: float = 0.3,
    final_div: float = None,
    wd: float = None,
    callbacks: Optional[CallbackList] = None,
    tot_epochs: int = None,
    start_epoch: int = None,
) -> None:
    "Fit a model following the 1cycle policy."
    max_lr = learn.lr_range(max_lr)
    callbacks = listify(callbacks)
    cyclic_scheduler = OneCycleScheduler(
        learn,
        max_lr,
        moms=moms,
        div_factor=div_factor,
        pct_start=pct_start,
        final_div=final_div,
        tot_epochs=tot_epochs,
        start_epoch=start_epoch,
    )
    callbacks.append(TFOneCycleScheduler(learn, scheduler=cyclic_scheduler))

    learn.fit(cyc_len, max_lr, wd=wd, callbacks=callbacks, one_cycle=True)


from fastai.basic_train import Recorder

TFRecorder.plot_losses = Recorder.plot_losses
TFRecorder._split_list = Recorder._split_list
TFRecorder._split_list_val = Recorder._split_list
TFRecorder.format_stats = Recorder.format_stats
EfficientDetLearner.fit_one_cycle = tf_fit_one_cycle
EfficientDetLearner.lr_find = efficient_lr_find



import os
import glob
import json
import random
import warnings
from shutil import ExecError
import traceback
from pathlib import Path

try:
    # TODO: remove unused imports
    import torch
    import cv2
    import numpy as np
    import matplotlib.pyplot as plt
    from numpy import mod
    from torch import resize_as_
    from fastai.vision import flatten_model
    from fastai.vision import ImageList
    from fastai.vision import Image
    from fastai.vision.image import open_image
    from fastai.vision.data import ImageDataBunch, ImageList
    from fastai.vision import imagenet_stats, normalize
    from fastai.vision.transform import crop, rotate, get_transforms, ResizeMethod

    from .._utils.common import _get_emd_path
    from ._arcgis_model import _EmptyData, _get_device
    from .._utils.env import raise_fastai_import_error
    from .._utils.pascal_voc_rectangles import ObjectDetectionCategoryList
    from ._siammask import Track  # TODO: need to move from Siammask to _tracking_utils
    from ._deepsort_train_utils import (
        get_learner,
        check_data_sanity,
        get_default_backbone,
        get_default_imgsize,
        load_for_prediction,
        get_fake_data,
    )
    from ._deepsort_predict_utils import DeepSortPredictor, get_corrected_labels_scores
    from ._tracker_util import TrackStatus

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

from ._arcgis_model import ArcGISModel
from .._utils.env import is_arcgispronotebook


class DeepSort(ArcGISModel):
    """
    Creates a :class:`~arcgis.learn.DeepSort` object.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function with `dataset_type=Imagenet`.
                            Default value is None.
                            DeepSort only supports image size of (3, 128, 64)
    =====================   ===========================================

    :return: :class:`~arcgis.learn.DeepSort` Object
    """

    # TODO: kwargs description

    def __init__(self, data, **kwargs):
        if not HAS_FASTAI:
            raise_fastai_import_error(import_exception=import_exception)

        if not check_data_sanity(data, self.supported_datasets):
            raise Exception("\nInvalid data format\n")

        if "backbone" not in kwargs:
            kwargs["backbone"] = None
        if data is None:
            data = get_fake_data()

        super().__init__(data, **kwargs)

        self._backend = "pytorch"
        self._is_multispectral = False
        self._num_classes = kwargs.get(
            "num_classes", len(self._data.class_mapping.items())
        )
        self._img_size = kwargs.get("img_size", get_default_imgsize())
        self._pretrained_path = kwargs.get("pretrained_path")
        self._infer_config = kwargs.get("infer_config", DeepSort._get_default_config())

        if not hasattr(self._data, "resize_to") or not self._data.resize_to:
            self._data.resize_to = get_default_imgsize()

        class ReID:
            def __init__(self, name=None):
                if name is None:
                    name = get_default_backbone()
                self.__name__ = name

            def get_name(self):
                return self.__name__

        self._backbone = ReID(kwargs.get("backbone"))
        self._device = _get_device()
        self.learn = get_learner(
            data, self._num_classes, self._backbone.get_name(), self._device
        )

        self._arcgis_init_callback()

        self.track_list = []
        self._predictor = None
        self._update_interval = 1
        if self._pretrained_path is not None:
            try:
                self.load(self._pretrained_path)
            except Exception:
                pass

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @classmethod
    def _get_default_label(cls):
        return "Object"

    @classmethod
    def get_model_type(cls):
        return "ObjectTracking"

    @classmethod
    def _get_default_config(cls):
        """
        Returns default config for inference.
        """
        infer_config = dict()
        infer_config["max_dist"] = 0.2
        infer_config["min_confidence"] = 0.1
        infer_config["nms_max_overlap"] = 0.9
        infer_config["max_iou_distance"] = 0.7
        infer_config["max_age"] = 70
        infer_config["n_init"] = 0
        infer_config["nn_budget"] = 100

        return infer_config

    @staticmethod
    def _supported_datasets():
        return ["Imagenet"]

    @staticmethod
    def _supported_backbones():
        return ["reid_v1", "reid_v2"]

    @property
    def supported_backbones(self):
        """Supported torchvision backbones for this model."""
        return DeepSort._supported_backbones()

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return DeepSort._supported_datasets()

    # TODO: bug crash on larger dataset, write custom show_results
    def show_results(self, rows=5):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        """
        self._check_requisites()
        from fastai.basic_data import DatasetType

        if self.learn.dl(DatasetType.Valid).batch_size > len(self.learn.data.valid_ds):
            rows = min(rows, len(self.learn.data.valid_ds))

        self.learn.show_results(rows=rows)
        if is_arcgispronotebook():
            plt.show()

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelType"] = DeepSort.get_model_type()
        _emd_template["BatchSize"] = self._data.batch_size
        _emd_template["NumClasses"] = len(self._data.classes)
        _emd_template["InferConfig"] = dict()

        for k, v in self._infer_config.items():
            _emd_template["InferConfig"][k] = v

        _emd_template["Classes"] = []
        inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
        class_data = {}
        for i, class_name in enumerate(self._data.classes):
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        return _emd_template

    def plot_confusion_matrix(self, **kwargs):
        self._check_requisites()
        from fastai.vision.learner import _cl_int_from_learner
        from fastai.vision.learner import ClassificationInterpretation

        import copy

        learn_temp = copy.copy(self.learn)

        # Reassigning the function from vision.learner because fastai sets it from tabular.learner
        ClassificationInterpretation.from_learner = _cl_int_from_learner
        interp = ClassificationInterpretation.from_learner(learn_temp)

        nrows = self._data.c
        # figsize range: 4 <= (no. of classes + 15)/4 <=20
        fs = min(max(4, (nrows + 15) / 4), 20)
        interp.plot_confusion_matrix(figsize=(fs, fs))

    def _save_confusion_matrix(self, path):
        from IPython.utils import io

        with io.capture_output() as captured:
            self.plot_confusion_matrix()
            plt.savefig(os.path.join(path, "confusion_matrix.png"))
            plt.close()

    @property
    def _model_metrics(self):
        return {
            # "accuracy": self.batch_accuracy(show_progress=True)
        }

    # TODO: option to pass infer-config
    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a DeepSort Object tracker from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.DeepSort` Object
        """

        if not HAS_FASTAI:
            raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        num_classes = int(emd["NumClasses"])
        pretrained_path = str(model_file)
        chip_size = emd["ImageWidth"]
        infer_config = dict()
        for k, v in emd["InferConfig"].items():
            infer_config[k] = v

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}

        if data is None:
            train_tfms = [rotate(degrees=30, p=0.5)]
            val_tfms = []
            transforms = (train_tfms, val_tfms)

            # TODO: possible duplication of code - see get_fake_data
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                kwargs_transforms = {}
                resize_to = emd.get("resize_to")
                kwargs_transforms["size"] = (resize_to[0], resize_to[1])
                kwargs_transforms["resize_method"] = ResizeMethod.SQUISH
                data = ImageDataBunch.single_from_classes(
                    emd_path.parent.parent,
                    sorted(list(class_mapping.values())),
                    ds_tfms=transforms,
                ).normalize(imagenet_stats)

            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.classes = list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data._dataset_type = "Imagenet"
        return cls(
            data,
            **model_params,
            pretrained_path=pretrained_path,
            num_classes=num_classes,
            infer_config=infer_config
        )

    def update(self, frame, detections=None, labels=None, scores=None, **kwargs):
        """
        Updates the :class:`~arcgis.learn.DeepSort` tracker.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. Frame is used to
                                update the tracker.
        ---------------------   -------------------------------------------
        detections              Required list. A list of bounding boxes
                                corresponding to the detections.
                                bounding box = [xmin, ymin, width, height]
        ---------------------   -------------------------------------------
        labels                  Optional list. A list of labels
                                corresponding to the detections.
        ---------------------   -------------------------------------------
        scores                  Optional list.  A list of scores
                                corresponding to the detections.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.Track` list
        """
        if detections is None:
            detections = []
        label = DeepSort._get_default_label()
        labels, scores = get_corrected_labels_scores(
            labels, scores, len(detections), label
        )
        bbox_xywh = np.asarray(detections)
        cls_names = np.asarray(labels)
        cls_conf = np.asarray(scores)

        if bbox_xywh is None or bbox_xywh.size == 0:
            bbox_xywh = np.array([]).reshape(0, 4)
            cls_conf = np.array([])
            labels = np.array([])

        outputs = np.array([])
        if self._predictor is not None and frame is not None:
            outputs, track_labels = self._predictor.update(
                bbox_xywh, cls_conf, cls_names, frame
            )
        else:
            print(
                "\nDeepSort predictor Not initialized. Please call init() before update()\n"
            )

        # TODO: if detect_interval == 1, check for track_status
        if np.size(outputs) > 0:
            bbox_xywh = outputs[:, :4].astype(np.int)
            track_identities = outputs[:, -3].astype(np.int)
            track_scores = outputs[:, -2]
            track_ages = outputs[:, -1].astype(np.int)
            for i in range(0, len(self.track_list)):
                self.track_list[i].status = TrackStatus.lost.value

            for idt in range(0, len(track_identities)):
                create_new_track = True
                current_track_idx = -1
                for i in range(0, len(self.track_list)):
                    track = self.track_list[i]
                    if track_identities[idt] == track.id:
                        current_track_idx = i
                        create_new_track = False
                        break

                if create_new_track is True:
                    track = Track(track_identities[idt], None, None, None)
                    self.track_list.append(track)
                    current_track_idx = len(self.track_list) - 1

                self.track_list[current_track_idx].bbox = bbox_xywh[idt]
                self.track_list[current_track_idx].label = track_labels[idt]
                self.track_list[current_track_idx].score = track_scores[idt]
                self.track_list[current_track_idx].age = track_ages[idt]
                self.track_list[current_track_idx].status = TrackStatus.tracking.value

        delete_indices = []
        for i in range(0, len(self.track_list)):
            if self.track_list[i].status == TrackStatus.lost.value:
                delete_indices.append(i)

        delete_indices = sorted(delete_indices, reverse=True)
        for idx in delete_indices:
            del self.track_list[idx]
        return self.track_list

    def init(self, frame, detections=None, labels=None, scores=None, **kwargs):
        """
        Initializes the :class:`~arcgis.learn.DeepSort` tracker for inference.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. Frame is used to
                                initialize the tracker.
        ---------------------   -------------------------------------------
        detections              Required list. A list of bounding boxes
                                corresponding to the detections.
        ---------------------   -------------------------------------------
        labels                  Optional list. A list of labels
                                corresponding to the detections.
        ---------------------   -------------------------------------------
        scores                  Optional list.  A list of scores
                                corresponding to the detections.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.Track` list
        """
        self.track_list = []
        self._update_interval = kwargs.get("update_interval", self._update_interval)
        deepsort = None
        learn_model = None
        if self._pretrained_path is None or not os.path.isfile(self._pretrained_path):
            if self.learn is not None:
                learn_model = self.learn.model
            else:
                raise Exception("\nInvalid initialization\n")

        try:
            deepsort = load_for_prediction(
                self._pretrained_path,
                self._num_classes,
                self._backbone.get_name(),
                learn_model,
            )
        except Exception:
            raise Exception("\nError loading model\n")

        if deepsort is not None:
            # TODO: separate predictor for each label
            self._predictor = DeepSortPredictor(
                deepsort, self._infer_config, self._device, self._update_interval
            )
        else:
            raise Exception("\nFailed Initialization\n")

        return self.update(frame, detections, labels, scores)

    # TODO: fix unchecked increment
    def remove(self, track_ids):
        """
        Removes the tracks from the track list using track_ids.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        track_ids               Required list. list of track ids to be removed
                                from the track list.
        =====================   ===========================================

        :return: Updated track list
        """
        try:
            for track_id in track_ids:
                for i in range(0, len(self.track_list)):
                    track = self.track_list[i]
                    if track.id == track_id:
                        del self.track_list[i]
                        break
            if self._predictor is not None:
                self._predictor.remove(track_ids)
        except Exception as e:
            print(e)

        return self.track_list



try:
    import tensorboardX
    from fastai.basics import *
    from fastai.vision import Learner
    from fastai.callbacks.tensorboard import LearnerTensorboardWriter
    from ..models._unet_utils import ArcGISSegmentationItemList
    from fastai.vision.data import ImageList, ImageImageList
    from torch.utils.tensorboard import SummaryWriter
    from fastai.callbacks.tensorboard import *
    from fastai.core import split_kwargs_by_func
    from PIL import Image
    from torchvision.transforms import ToTensor
except:
    pass


class ArcGISTBCallback(
    LearnerTensorboardWriter,
    Learner,
    ImageImageList,
    ArcGISSegmentationItemList,
    ImageList,
):
    def __init__(self, learn, base_dir, name, arcgis_model):
        self._base_dir = base_dir
        self._name = name
        self._arcgis_model = arcgis_model
        self._current_epoch = 0
        self._current_run = name + str(self._current_epoch)
        super(ArcGISTBCallback, self).__init__(learn, base_dir, self._current_run)

    def on_train_begin(self, **kwargs: Any):
        pass

    # Override the on_train_begin method of the parent class as it causes graph related errors and warnings.#5244

    def on_epoch_end(self, last_metrics: MetricsList, iteration: int, **kwargs) -> None:
        self._current_epoch = self._current_epoch + 1
        self._current_run = self._name + "-Epoch-" + str(self._current_epoch)
        log_dir = self._base_dir / self._current_run
        self.tbwriter = SummaryWriter(str(log_dir))
        "Callback function that writes epoch end appropriate data to Tensorboard."
        if not all(x is None for x in last_metrics):
            self._write_metrics(iteration=iteration, last_metrics=last_metrics)
        # self._write_embedding(iteration=iteration)
        self._write_figure(iteration=iteration)

    def _write_figure(self, iteration: int) -> None:
        tag = "predictions vs. actuals"
        rows = 4
        thresh = 0.1
        nms_overlap = 0.1
        obj_det_models = ["FeatureClassifier", "SingleShotDetector", "RetinaNet"]
        img_to_img_models = [
            "UnetClassifier",
            "SuperResolution",
            "PSPNetClassifier",
            "DeepLab",
        ]
        other_models = [
            "ImageCaptioner",
            "MaskRCNN",
            "MultiTaskRoadExtractor",
            "ConnectNet",
        ]
        text_models = ["TextClassifier"]
        if (type(self._arcgis_model).__name__) in obj_det_models:
            fig1 = self.show_results(
                rows=rows,
                thresh=thresh,
                nms_overlap=nms_overlap,
                model=self._arcgis_model,
            )
        elif (type(self._arcgis_model).__name__) in img_to_img_models:
            if (type(self._arcgis_model).__name__) == "SuperResolution":
                if self.data.train_ds.__class__.__name__ == "SR3Dataset":
                    rows = 1
                else:
                    rows = rows
                fig1 = self.show_results_superres(rows=rows)
            else:
                fig1 = self.show_results(rows=rows)  # Segmentation
        elif (type(self._arcgis_model).__name__) in other_models:
            fig1 = self._arcgis_model.show_results(2, return_fig=True)
        # elif (type(self._arcgis_model).__name__) in text_models:
        # txt= self._arcgis_model.show_results(return_text=True)
        elif (type(self._arcgis_model).__name__) == "FasterRCNN":
            if self.data._is_multispectral:
                fig1 = self._arcgis_model._show_results_multispectral(return_fig=True)
            else:
                fig1 = self._show_results_modified(2, return_fig=True)
        elif (type(self._arcgis_model).__name__) == "CycleGAN":
            self._arcgis_model.learn.model.arcgis_results = True
            fig1 = self.show_results(rows=rows)
            self._arcgis_model.learn.model.arcgis_results = False
        elif (type(self._arcgis_model).__name__) == "Pix2Pix" or (
            type(self._arcgis_model).__name__
        ) == "Pix2PixHD":
            fig1 = self.show_results_pix2pix(rows=rows)
        else:
            return

        # Removing the support for Point CNN due to plotly orca dependency
        # if (type(self._arcgis_model).__name__) == 'PointCNN':
        #    fig1.write_image("fig1_pcnn.jpeg")
        #    image = Image.open("fig1_pcnn.jpeg")
        #    image = ToTensor()(image)
        #    self.tbwriter.add_image(tag=tag, img_tensor=image, global_step=iteration)
        # elif (type(self._arcgis_model).__name__) in text_models:
        # self.tbwriter.add_text(tag=tag, text_string =txt, global_step=iteration)
        # else:
        self.tbwriter.add_figure(
            tag=tag, figure=fig1, global_step=iteration, close=True
        )

    def show_results(self, ds_type=DatasetType.Valid, rows: int = 5, **kwargs):
        "Show `rows` result of predictions on `ds_type` dataset."
        n_items = rows**2 if self.data.train_ds.x._square_show_res else rows
        if self.dl(ds_type).batch_size < n_items:
            n_items = self.dl(ds_type).batch_size
        ds = self.dl(ds_type).dataset
        self.callbacks.append(RecordOnCPU())
        preds = self.pred_batch(ds_type)
        *self.callbacks, rec_cpu = self.callbacks
        x, y = rec_cpu.input, rec_cpu.target
        norm = getattr(self.data, "norm", False)
        if norm:
            x = self.data.denorm(x)
            if norm.keywords.get("do_y", False):
                y = self.data.denorm(y, do_x=True)
                preds = self.data.denorm(preds, do_x=True)
        analyze_kwargs, kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        if n_items > len(preds):
            n_items = len(preds)
        preds = [
            ds.y.analyze_pred(grab_idx(preds, i), **analyze_kwargs)
            for i in range(n_items)
        ]
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(n_items)]
        if has_arg(ds.y.reconstruct, "x"):
            ys = [ds.y.reconstruct(grab_idx(y, i), x=x) for i, x in enumerate(xs)]
            zs = [ds.y.reconstruct(z, x=x) for z, x in zip(preds, xs)]
        else:
            ys = [ds.y.reconstruct(grab_idx(y, i)) for i in range(n_items)]
            zs = [ds.y.reconstruct(z) for z in preds]
        if (
            (type(self._arcgis_model).__name__) == "FeatureClassifier"
            or (type(self._arcgis_model).__name__) == "SingleShotDetector"
            or (type(self._arcgis_model).__name__) == "RetinaNet"
        ):
            if self.data._is_multispectral:
                fig1 = self._arcgis_model._show_results_multispectral(return_fig=True)
            else:
                fig1 = self.show_xyzs(xs, ys, zs, **kwargs)
        elif (
            (type(self._arcgis_model).__name__) == "UnetClassifier"
            or (type(self._arcgis_model).__name__) == "PSPNetClassifier"
            or (type(self._arcgis_model).__name__) == "DeepLab"
        ):
            if self.data._is_multispectral:
                fig1 = self._arcgis_model._show_results_multispectral(return_fig=True)
            else:
                fig1 = self.segment_show_xyzs(xs, ys, zs)
        elif (type(self._arcgis_model).__name__) == "SuperResolution":
            fig1 = self.img_img_show_xyzs(xs, ys, zs)
        elif (type(self._arcgis_model).__name__) == "CycleGAN":
            fig1 = self.img_tuple_show_xyzs(xs, ys, zs)
        return fig1

    def show_xyzs(
        self,
        xs,
        ys,
        zs,
        imgsize: int = 4,
        figsize: Optional[Tuple[int, int]] = None,
        **kwargs,
    ):
        "Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`."
        if (
            (type(self._arcgis_model).__name__) == "SingleShotDetector"
            or (type(self._arcgis_model).__name__) == "RetinaNet"
            or (type(self._arcgis_model).__name__) == "FasterRCNN"
        ):
            self._square_show_res = False
        else:
            self._square_show_res = True
        if self._square_show_res:
            title = "Ground truth\nPredictions"
            rows = int(np.ceil(math.sqrt(len(xs))))
            fig, axs = self.subplots(
                rows,
                rows,
                imgsize=imgsize,
                figsize=figsize,
                title=title,
                weight="bold",
                size=12,
            )
            for x, y, z, ax in zip(xs, ys, zs, axs.flatten()):
                x.show(ax=ax, title=f"{str(y)}\n{str(z)}", **kwargs)
            for ax in axs.flatten()[len(xs) :]:
                ax.axis("off")
        else:
            title = "Ground truth/Predictions"
            fig, axs = self.subplots(
                len(xs),
                2,
                imgsize=imgsize,
                figsize=figsize,
                title=title,
                weight="bold",
                size=14,
            )
            for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
                x.show(ax=axs[i, 0], y=y, **kwargs)
                x.show(ax=axs[i, 1], y=z, **kwargs)

        return fig

    def show_xys(
        self,
        xs,
        ys,
        imgsize: int = 4,
        figsize: Optional[Tuple[int, int]] = None,
        **kwargs,
    ):
        "Show the `xs` (inputs) and `ys`(targets)  on a figure of `figsize`."
        fig, axs = self.subplots(len(xs), 2, imgsize=imgsize, figsize=figsize)
        for i, (x, y) in enumerate(zip(xs, ys)):
            x.show(ax=axs[i, 0], **kwargs)
            y.show(ax=axs[i, 1], **kwargs)
        plt.tight_layout()
        return fig

    def subplots(
        self,
        rows: int,
        cols: int,
        imgsize: int = 4,
        figsize: Optional[Tuple[int, int]] = None,
        title=None,
        **kwargs,
    ):
        "Like `plt.subplots` but with consistent axs shape, `kwargs` passed to `fig.suptitle` with `title`"
        figsize = ifnone(figsize, (imgsize * cols, imgsize * rows))
        fig, axs = plt.subplots(rows, cols, figsize=figsize)
        if rows == cols == 1:
            axs = [[axs]]  # subplots(1,1) returns Axes, not [Axes]
        elif (rows == 1 and cols != 1) or (cols == 1 and rows != 1):
            axs = [axs]
        if title is not None:
            fig.suptitle(title, **kwargs)
        return fig, array(axs)

    def segment_show_xyzs(
        self,
        xs,
        ys,
        zs,
        imgsize: int = 4,
        figsize: Optional[Tuple[int, int]] = None,
        **kwargs,
    ):
        "Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`."
        if self._square_show_res:
            title = "Ground truth\nPredictions"
            rows = int(np.ceil(math.sqrt(len(xs))))
            fig, axs = self.subplots(
                rows,
                rows,
                imgsize=imgsize,
                figsize=figsize,
                title=title,
                weight="bold",
                size=12,
            )
            for x, y, z, ax in zip(xs, ys, zs, axs.flatten()):
                x.show(ax=ax, title=f"{str(y)}\n{str(z)}", **kwargs)
            for ax in axs.flatten()[len(xs) :]:
                ax.axis("off")
        else:
            title = "Ground truth/Predictions"
            fig, axs = self.subplots(
                len(xs),
                2,
                imgsize=imgsize,
                figsize=figsize,
                title=title,
                weight="bold",
                size=14,
            )
            for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
                x.show(ax=axs[i, 0], y=y, **kwargs)
                x.show(ax=axs[i, 1], y=z, **kwargs)
        return fig

    def show_results_superres(self, rows, **kwargs):
        from .._data_utils.pix2pix_data import display_row, denormalize
        from .._utils.common import get_nbatches, get_top_padding
        from .common import ArcGISMSImage
        from math import ceil

        sampling = kwargs.get("sampling_type", "ddim")
        ntimestep = kwargs.get("n_timestep", None)
        device = next(self.learn.model.parameters()).device.type

        self.learn.model.eval()
        activ = []
        top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
        denormfunc = lambda img, max, min: (img + 1) * (max - min) / 2 + min

        if self._arcgis_model.model_type == "UNet":
            x_batch, y_batch = get_nbatches(
                self._arcgis_model._data.valid_dl,
                ceil(rows / self._arcgis_model._data.batch_size),
            )
            if isinstance(x_batch[0], list):
                x_batch = x_batch[0]

            x_A, x_B = torch.cat(x_batch), torch.cat(y_batch)

            for i in range(0, x_A.shape[0], self._arcgis_model._data.batch_size):
                preds = self._arcgis_model.learn.model(
                    x_A[i : i + self._arcgis_model._data.batch_size].detach()
                )
                activ.append(preds)
            activations = torch.cat(activ)

            x_A = denormalize(x_A.cpu(), *self._arcgis_model._data._image_stats)
            x_B = denormalize(x_B.cpu(), *self._arcgis_model._data._image_stats)
            activations = denormalize(
                activations.cpu(), *self._arcgis_model._data._image_stats
            )
            rows = min(rows, x_A.shape[0])
        else:
            x_A = torch.cat([i[0] for i, _ in self._arcgis_model.learn.data.valid_dl])[
                :rows
            ]
            x_B = torch.cat([j for _, j in self._arcgis_model.learn.data.valid_dl])[
                :rows
            ]

            x_A_batch = x_A.detach()

            if sampling == "ddim":
                n_timestep = ntimestep if ntimestep else 200
                nstp = {"n_timestep": self._arcgis_model.kwargs.get("n_timestep", 1000)}
            else:
                n_timestep = (
                    ntimestep
                    if ntimestep
                    else self._arcgis_model.kwargs.get("n_timestep", 1000)
                )
                nstp = {"n_timestep": n_timestep}
            combkwargs = {**kwargs, **self._arcgis_model.kwargs, **nstp}
            self._arcgis_model.learn.model.set_new_noise_schedule(device, **combkwargs)

            preds = []
            for k in range(x_A_batch.shape[0]):
                if sampling == "ddim":
                    preds.append(
                        self._arcgis_model.learn.model.super_resolution(
                            x_A_batch[k, None],
                            continous=False,
                            sampling_timesteps=n_timestep,
                            ddim_sampling_eta=1,
                            sampling="ddim",
                        )
                    )
                else:
                    preds.append(
                        self._arcgis_model.learn.model.super_resolution(
                            x_A_batch[k, None], continous=False
                        )
                    )
            activations = torch.cat(preds)

            maxvals_a = self._arcgis_model._data.batch_stats_a["band_max_values"][
                ..., None, None
            ]
            minvals_a = self._arcgis_model._data.batch_stats_a["band_min_values"][
                ..., None, None
            ]
            maxvals_b = self._arcgis_model._data.batch_stats_b["band_max_values"][
                ..., None, None
            ]
            minvals_b = self._arcgis_model._data.batch_stats_b["band_min_values"][
                ..., None, None
            ]
            x_A = denormfunc(x_A.cpu(), maxvals_a, minvals_a)
            x_B = denormfunc(x_B.cpu(), maxvals_b, minvals_b)
            activations = denormfunc(activations.cpu(), maxvals_b, minvals_b)

        fig, axs = plt.subplots(
            nrows=rows, ncols=3, figsize=(4 * 5, rows * 5), squeeze=False
        )
        plt.subplots_adjust(top=top)
        axs[0, 0].title.set_text("Input")
        axs[0, 1].title.set_text("Target")
        axs[0, 2].title.set_text("Prediction")
        for r in range(rows):
            display_row(
                axs[r],
                (
                    ArcGISMSImage(x_A[r].cpu()),
                    ArcGISMSImage(x_B[r].cpu()),
                    ArcGISMSImage(activations[r].detach().cpu()),
                ),
                kwargs.get("rgb_bands", None),
            )
        return fig

    def img_img_show_xyzs(
        self,
        xs,
        ys,
        zs,
        imgsize: int = 4,
        figsize: Optional[Tuple[int, int]] = None,
        **kwargs,
    ):
        if (type(self._arcgis_model).__name__) == "SuperResolution":
            self._square_show_res = False
            self._square_show = False
        "Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`."
        title = "Input / Prediction / Target"
        fig, axs = self.subplots(
            len(xs),
            3,
            imgsize=imgsize,
            figsize=figsize,
            title=title,
            weight="bold",
            size=14,
        )
        for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
            x.show(ax=axs[i, 0], **kwargs)
            y.show(ax=axs[i, 2], **kwargs)
            z.show(ax=axs[i, 1], **kwargs)
        return fig

    def img_tuple_show_xyzs(
        self, xs, ys, zs, figsize: Tuple[int, int] = None, **kwargs
    ):
        """Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`.
        `kwargs` are passed to the show method."""
        figsize = ifnone(figsize, (12, 3 * len(xs)))
        fig, axs = plt.subplots(len(xs), 2, figsize=figsize)
        fig.suptitle("Ground truth / Predictions", weight="bold", size=14)
        for i, (x, z) in enumerate(zip(xs, zs)):
            x.to_one().show(ax=axs[i, 0], **kwargs)
            z.to_one_pred().show(ax=axs[i, 1], **kwargs)
        return fig

    def _show_results_modified(self, rows=5, **kwargs):
        if rows > len(self._arcgis_model._data.valid_ds):
            rows = len(self._arcgis_model._data.valid_ds)

        ds_type = DatasetType.Valid
        xb, yb = self._arcgis_model.learn.data.one_batch(
            ds_type, detach=False, denorm=False
        )
        ds = self._arcgis_model.learn.dl(ds_type).dataset
        if xb.shape[0] < rows**2:
            ds_type = DatasetType.Train
            xb, yb = self._arcgis_model.learn.data.one_batch(
                ds_type, detach=False, denorm=False
            )
            ds = self._arcgis_model.learn.dl(ds_type).dataset
        n_items = (
            rows**2
            if self._arcgis_model.learn.data.train_ds.x._square_show_res
            else rows
        )
        if self._arcgis_model.learn.dl(ds_type).batch_size < n_items:
            n_items = self._arcgis_model.learn.dl(ds_type).batch_size
        self._arcgis_model.learn.model.eval()
        transform_kwargs, kwargs = split_kwargs_by_func(
            kwargs, self._arcgis_model._model_conf.transform_input
        )
        try:
            preds = self._arcgis_model.learn.model(
                self._arcgis_model._model_conf.transform_input(xb, transform_kwargs)
            )
        except Exception as e:
            if getattr(self._arcgis_model, "_is_fasterrcnn", False):
                preds = []
                for _ in range(xb.shape[0]):
                    res = {}
                    res["boxes"] = torch.empty(0, 4)
                    res["scores"] = torch.tensor([])
                    res["labels"] = torch.tensor([])
                    preds.append(res)
            else:
                raise e

        x, y = to_cpu(xb), to_cpu(yb)
        norm = getattr(self._arcgis_model.learn.data, "norm", False)
        if norm:
            x = self._arcgis_model.learn.data.denorm(x)
            if norm.keywords.get("do_y", False):
                y = self._arcgis_model.learn.data.denorm(y, do_x=True)
                preds = self._arcgis_model.learn.data.denorm(preds, do_x=True)
        analyze_kwargs, kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        preds = ds.y.analyze_pred(preds, self._arcgis_model, **analyze_kwargs)
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(n_items)]
        if has_arg(ds.y.reconstruct, "x"):
            ys = [ds.y.reconstruct(grab_idx(y, i), x=x) for i, x in enumerate(xs)]
            zs = [ds.y.reconstruct(z, x=x) for z, x in zip(preds, xs)]
        else:
            ys = [ds.y.reconstruct(grab_idx(y, i)) for i in range(n_items)]
            zs = [ds.y.reconstruct(z) for z in preds]
        fig1 = self.show_xyzs(xs, ys, zs, **kwargs)
        return fig1

    def show_results_pix2pix(self, rows, **kwargs):
        from .common import get_nbatches, get_top_padding, ArcGISMSImage
        from .._data_utils.pix2pix_data import denormalize, display_row

        self._arcgis_model.learn.model.eval()
        x_batch, y_batch = get_nbatches(
            self._arcgis_model._data.valid_dl,
            math.ceil(rows / self._arcgis_model._data.batch_size),
        )
        x_A, x_B = [x[0] for x in x_batch], [x[1] for x in x_batch]

        x_A = torch.cat(x_A)
        x_B = torch.cat(x_B)

        top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
        activ = []

        for i in range(0, x_B.shape[0], self._arcgis_model._data.batch_size):
            with torch.no_grad():
                preds = self._arcgis_model.learn.model(
                    x_A[i : i + self._arcgis_model._data.batch_size].detach(),
                    x_B[i : i + self._arcgis_model._data.batch_size].detach(),
                )
            activ.append(preds[0])

        activations = torch.cat(activ)
        if self._arcgis_model._data.label_nc == 0:
            x_A = denormalize(x_A.cpu(), *self._arcgis_model._data.norm_stats)
        x_B = denormalize(x_B.cpu(), *self._arcgis_model._data.norm_stats)
        activations = denormalize(
            activations.cpu(), *self._arcgis_model._data.norm_stats
        )
        rows = min(rows, x_A.shape[0])

        fig, axs = plt.subplots(
            nrows=rows, ncols=3, figsize=(4 * 5, rows * 5), squeeze=False
        )
        plt.subplots_adjust(top=top)
        axs[0, 0].title.set_text("Input")
        axs[0, 1].title.set_text("Ground Truth")
        axs[0, 2].title.set_text("Prediction")
        from fastai.vision import image2np

        for r in range(rows):
            if self._arcgis_model._data._is_multispectral:
                display_row(
                    axs[r],
                    (
                        ArcGISMSImage(x_A[r]),
                        ArcGISMSImage(x_B[r]),
                        ArcGISMSImage(activations[r]),
                    ),
                )
            else:
                display_row(
                    axs[r],
                    (image2np(x_A[r]), image2np(x_B[r]), image2np(activations[r])),
                )
        return fig



import torch
from scipy.optimize import linear_sum_assignment
from torch import nn
from arcgis.learn._utils.coco_detection_utils import box_cxcywh_to_xyxy
import numpy as np
from torchvision.ops.boxes import box_area


def box_iou(boxes1, boxes2):
    area1 = box_area(boxes1)
    area2 = box_area(boxes2)

    lt = torch.max(boxes1[..., :2], boxes2[..., :2])  # [N,M,2]
    rb = torch.min(boxes1[..., 2:], boxes2[..., 2:])  # [N,M,2]

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    inter = wh[..., 0] * wh[..., 1]  # [N,M]

    union = area1 + area2 - inter

    iou = inter / union
    return iou, union


def generalized_box_iou(boxes1, boxes2):
    """
    Generalized IoU from https://giou.stanford.edu/

    The boxes should be in [x0, y0, x1, y1] format

    Returns a [N, M] pairwise matrix, where N = len(boxes1)
    and M = len(boxes2)
    """
    # degenerate boxes gives inf / nan results
    # so do an early check
    assert (boxes1[..., 2:] >= boxes1[..., :2]).all()
    assert (boxes2[..., 2:] >= boxes2[..., :2]).all()
    iou, union = box_iou(boxes1, boxes2)

    lt = torch.min(boxes1[..., :2], boxes2[..., :2])
    rb = torch.max(boxes1[..., 2:], boxes2[..., 2:])

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    area = wh[..., 0] * wh[..., 1]

    return iou - (area - union) / area


def return_matching(cost):
    indices = np.array(list(map(linear_sum_assignment, cost.detach().cpu().numpy())))
    indices = np.transpose(indices, axes=(0, 2, 1))[..., 1:]
    indices = torch.LongTensor(indices).to(cost).type(torch.int64)
    return indices


class HungarianMatcher(nn.Module):
    """This class computes an assignment between the targets and the predictions of the network

    For efficiency reasons, the targets don't include the no_object. Because of this, in general,
    there are more predictions than targets. In this case, we do a 1-to-1 matching of the best predictions,
    while the others are un-matched (and thus treated as non-objects).
    """

    def __init__(
        self, cost_class: float = 1, cost_bbox: float = 1, cost_giou: float = 1
    ):
        """Creates the matcher

        Params:
            cost_class: This is the relative weight of the classification error in the matching cost
            cost_bbox: This is the relative weight of the L1 error of the bounding box coordinates in the matching cost
            cost_giou: This is the relative weight of the giou loss of the bounding box in the matching cost
        """
        super().__init__()
        self.cost_class = cost_class
        self.cost_bbox = cost_bbox
        self.cost_giou = cost_giou
        assert (
            cost_class != 0 or cost_bbox != 0 or cost_giou != 0
        ), "all costs cant be 0"

    def forward(self, out_bbox, tgt_bbox):
        """Performs the matching

        Params:
            outputs: This is a dict that contains at least these entries:
                 "pred_logits": Tensor of dim [batch_size, num_queries, num_classes] with the classification logits
                 "pred_boxes": Tensor of dim [batch_size, num_queries, 4] with the predicted box coordinates

            targets: This is a list of targets (len(targets) = batch_size), where each target is a dict containing:
                 "labels": Tensor of dim [num_target_boxes] (where num_target_boxes is the number of ground-truth
                           objects in the target) containing the class labels
                 "boxes": Tensor of dim [num_target_boxes, 4] containing the target box coordinates

        Returns:
            A list of size batch_size, containing tuples of (index_i, index_j) where:
                - index_i is the indices of the selected predictions (in order)
                - index_j is the indices of the corresponding selected targets (in order)
            For each batch element, it holds:
                len(index_i) = len(index_j) = min(num_queries, num_target_boxes)
        """
        with torch.no_grad():
            bs, num_queries1 = out_bbox.shape[:2]
            _, num_queries2 = tgt_bbox.shape[:2]
            # We flatten to compute the cost matrices in a batch
            # out_prob = outputs["pred_logits"].flatten(0, 1).softmax(-1)  # [batch_size * num_queries, num_classes]
            # out_bbox = outputs["pred_boxes"].flatten(0, 1)  # [batch_size * num_queries, 4]
            # out_bbox = out_bbox.flatten(0, 1)
            # Also concat the target labels and boxes
            # tgt_ids = torch.cat([v["labels"] for v in targets])
            # tgt_bbox = tgt_bbox.flatten(0, 1)
            # Compute the classification cost. Contrary to the loss, we don't use the NLL,
            # but approximate it in 1 - proba[target class].
            # The 1 is a constant that doesn't change the matching, it can be ommitted.
            # cost_class = -out_prob[:, tgt_ids]

            # Compute the L1 cost between boxes
            out_bbox1 = out_bbox.unsqueeze(2).repeat_interleave(num_queries2, 2)
            tgt_bbox2 = tgt_bbox.unsqueeze(1).repeat_interleave(num_queries1, 1)
            cost_bbox = torch.nn.functional.l1_loss(
                out_bbox1, tgt_bbox2, reduction="none"
            ).mean(dim=-1)
            cost_giou = -generalized_box_iou(
                box_cxcywh_to_xyxy(out_bbox1.flatten(0, 2)),
                box_cxcywh_to_xyxy(tgt_bbox2.flatten(0, 2)),
            )
            cost_giou = cost_giou.view(bs, num_queries1, num_queries2)

            # Final cost matrix
            C = self.cost_bbox * cost_bbox + self.cost_giou * cost_giou
            indices = return_matching(C)
            # identity matching
            indices = (
                torch.arange(0, C.shape[1])
                .view(1, C.shape[1], 1)
                .expand_as(indices)
                .to(indices)
            )
        return indices


def build_matcher(args):
    return HungarianMatcher(
        cost_class=args.set_cost_class,
        cost_bbox=args.set_cost_bbox,
        cost_giou=args.set_cost_giou,
    )



"""
Tools to assist users to work with PKI Certificates
"""
import os
import tempfile
import cryptography
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PrivateFormat,
    NoEncryption,
)

_crypto_version = [
    int(i) if i.isdigit() else i for i in cryptography.__version__.split(".")
]


# ----------------------------------------------------------------------
def pfx_to_pem(pfx_path, pfx_password, folder=None, use_openssl=False):
    """Decrypts the .pfx file to be used with requests.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    pfx_path            Required string.  File pathname to .pfx file to parse.
    ---------------     --------------------------------------------------------------------
    pfx_password        Required string.  Password to open .pfx file to extract key/cert.
    ---------------     --------------------------------------------------------------------
    folder              Optional String.  The save location of the certificate files.  The
                        default is the tempfile.gettempdir() directory.
    ---------------     --------------------------------------------------------------------
    user_openssl        Optional Boolean. If True, OpenPySSL is used to convert the pfx to pem instead of cryptography.
    ===============     ====================================================================

    :return: Tuple
       File path to key_file located in a tempfile location
       File path to cert_file located in a tempfile location
    """
    if folder is None:
        folder = tempfile.gettempdir()
    elif folder and not os.path.isdir(folder):
        raise Exception("Folder location does not exist.")
    key_file = tempfile.NamedTemporaryFile(suffix=".pem", delete=False, dir=folder)
    cert_file = tempfile.NamedTemporaryFile(suffix=".pem", delete=False, dir=folder)
    if use_openssl:
        try:
            import OpenSSL.crypto

            k = open(key_file.name, "wb")
            c = open(cert_file.name, "wb")
            try:
                pfx = open(pfx_path, "rb").read()
                p12 = OpenSSL.crypto.load_pkcs12(pfx, pfx_password)
            except OpenSSL.crypto.Error:
                raise RuntimeError("Invalid PFX password.  Unable to parse file.")
            k.write(
                OpenSSL.crypto.dump_privatekey(
                    OpenSSL.crypto.FILETYPE_PEM, p12.get_privatekey()
                )
            )
            c.write(
                OpenSSL.crypto.dump_certificate(
                    OpenSSL.crypto.FILETYPE_PEM, p12.get_certificate()
                )
            )
            k.close()
            c.close()
        except ImportError as e:
            raise e
        except Exception as ex:
            raise ex
    else:
        _default_backend = None
        if _crypto_version < [3, 0]:
            from cryptography.hazmat.backends import default_backend

            _default_backend = default_backend()

        with open(pfx_path, "rb") as f:
            (
                private_key,
                certificate,
                additional_certificates,
            ) = pkcs12.load_key_and_certificates(
                f.read(), str.encode(pfx_password), backend=_default_backend
            )
        cert_bytes = certificate.public_bytes(Encoding.PEM)
        pk_bytes = private_key.private_bytes(
            Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()
        )
        k = open(key_file.name, "wb")
        c = open(cert_file.name, "wb")

        k.write(pk_bytes)
        c.write(cert_bytes)
        k.close()
        c.close()
        del k
        del c
    key_file.close()
    cert_file.close()
    return cert_file.name, key_file.name  # certificate/key



try:
    import pandas as pd
    from arcgis.features.geo._accessor import GeoAccessor
    from arcgis.features.geo._accessor import GeoSeriesAccessor
    from arcgis.features.geo._accessor import _is_geoenabled

    __all__ = ["GeoAccessor", "GeoSeriesAccessor"]
except ImportError:
    pass



from typing import Union, Optional, ClassVar
from dataclasses import field, dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.geometry import (
    _HasGeometry,
    SingleFieldGeometry,
    XYZGeometry,
)
from arcgis.realtime.velocity.feeds.time import _HasTime, TimeInstant, TimeInterval
from arcgis.realtime.velocity.input.format import (
    EsriJsonFormat,
    GeoJsonFormat,
    JsonFormat,
    DelimitedFormat,
    XMLFormat,
    _format_from_config,
)


@dataclass
class RabbitMQ(_FeedTemplate, _HasTime, _HasGeometry):
    """
    Receive events from a RabbitMQ broker. This data class can be used to define the feed configuration and
    to create the feed.

    ==================      ====================================================================
    **Parameter**            **Description**
    ------------------      --------------------------------------------------------------------
    label                   String. Unique label for the feed instance.
    ------------------      --------------------------------------------------------------------
    description             String. Feed description.
    ------------------      --------------------------------------------------------------------
    host                    String. Host address of the RabbitMQ Server.

                            For example:

                                rabbitmqbroker.centralus.cloudapp.azure.com
    ------------------      --------------------------------------------------------------------
    port                    int. Port on which the RabbitMQ Server is accessible. The default is:
                            5672.
    ------------------      --------------------------------------------------------------------
    use_ssl                 bool. Whether or not to use SSL in the connection. The default is:
                            False.
    ------------------      --------------------------------------------------------------------
    prefetch_count          int. Prefetch count is used to specify the number of messages
                            RabbitMQ sends. This limits how many messages are received before
                            acknowledging a message. The default is: 0.
    ==================      ====================================================================

    =====================   ====================================================================
    **Optional Argument**   **Description**
    =====================   ====================================================================
    virtual_host            String. Virtual host of the RabbitMQ Server.
                            For example: virtualhost1
    ---------------------   --------------------------------------------------------------------
    username                String. Username for server authentication.
    ---------------------   --------------------------------------------------------------------
    password                String. Password for server authentication.
    ---------------------   --------------------------------------------------------------------
    queue_name              String. Name of the queue over which messages will be received.
    ---------------------   --------------------------------------------------------------------
    data_format             [:class:`~arcgis.realtime.velocity.input.EsriJsonFormat`, :class:`~arcgis.realtime.velocity.input.GeoJsonFormat`, :class:`~arcgis.realtime.velocity.input.DelimitedFormat`, :class:`~arcgis.realtime.velocity.input.JsonFormat`, :class:`~arcgis.realtime.velocity.input.XMLFormat`]. An instance that contains the data format configuration
                            for this feed. Configure only allowed formats. If this is not set
                            right during initialization, a format will be auto-detected and set
                            from a sample of the incoming data. This sample will be fetched from
                            the configuration provided so far in the init.
    ---------------------   --------------------------------------------------------------------
    track_id_field          String. name of the field from the incoming data that should be set as
                            track ID.
    ---------------------   --------------------------------------------------------------------
    geometry                [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`]. An instance of geometry
                            configuration that will be used to create geometry objects from the
                            incoming data.
    ---------------------   --------------------------------------------------------------------
    time                    [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration
                            that will be used to create time information from the incoming data.
    =====================   ====================================================================

    :return: A data class with RabbitMQ feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import RabbitMQ
        from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant

        rabbitmq_config = RabbitMQ(
            label="feed_name",
            description="feed_description",
            host="RabbitMQ host",
            username="RabbitMQ_username",
            password="RabbitMQ password",
            prefetch_count=0,
            queue_name="RabbitMQ_queue_name",
            data_format=None
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        rabbitmq_feed = feeds.create(rabbitmq_config)
        rabbitmq_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # RabbitMQ specific properties
    host: str
    port: int = field(default=5672)
    use_ssl: bool = field(default=False)
    prefetch_count: int = field(default=0)
    virtual_host: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    queue_name: Optional[str] = None

    # user can define these properties even after initialization
    data_format: Optional[
        Union[DelimitedFormat, EsriJsonFormat, GeoJsonFormat, JsonFormat, XMLFormat]
    ] = None
    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasGeometry properties
    geometry: Optional[Union[XYZGeometry, SingleFieldGeometry]] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None

    # FeedTemplate properties
    _name: ClassVar[str] = "rabbitmq"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        test_connection = self._util.test_connection(
            input_type="feed", payload=feed_properties
        )
        # Test connection to make sure feed can fetch schema
        if test_connection is True:
            # test connection succeeded. Now try getting sample messages
            sample_payload = {
                "properties": {
                    "maxSamplesToCollect": 5,
                    "timeoutInMillis": 5000,
                }
            }
            self._dict_deep_merge(sample_payload, feed_properties)
            # Sample messages to fetch schema/fields
            sample_messages = self._util.sample_messages(
                input_type="feed", payload=sample_payload
            )

            if sample_messages["featureSchema"] is not None:
                # sample messages succeeded. Get Feature Schema from it
                self._set_fields(sample_messages["featureSchema"])

            # if Format was not specified by user, use the auto-detected format from the sample messages response as this feed object's format.
            if self.data_format is None:
                self.data_format = _format_from_config(sample_messages)

            print(
                "Feature Schema retrieved from the Feed:",
                sample_messages["featureSchema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.geometry is not None:
                self.set_geometry_config(self.geometry)
            if self.time is not None:
                self.set_time_config(self.time)

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        if self.virtual_host:
            virtual_host_prop = {f"{self._name}.virtualHost": self.virtual_host}
        else:
            virtual_host_prop = {}

        if self.username:
            username_prop = {f"{self._name}.username": self.username}
        else:
            username_prop = {}

        if self.password:
            password_prop = {f"{self._name}.password": self.password}
        else:
            password_prop = {}

        if self.queue_name:
            queue_name_prop = {f"{self._name}.queueName": self.queue_name}
        else:
            queue_name_prop = {}

        feed_properties = {
            "name": self._name,
            "properties": {
                f"{self._name}.host": self.host,
                f"{self._name}.port": self.port,
                f"{self._name}.useSSL": self.use_ssl,
                f"{self._name}.prefetchCount": self.prefetch_count,
                **virtual_host_prop,
                **username_prop,
                **password_prop,
                **queue_name_prop,
            },
        }

        if self.data_format is not None:
            format_dict = self.data_format._build()
            self._dict_deep_merge(feed_properties, format_dict)

        return feed_properties



from logging import exception
from ._arcgis_model import ArcGISModel, _EmptyData
from .._data import _raise_fastai_import_error
from ._codetemplate import image_translation_prf
import traceback, json

try:
    from ._pix2pix_hd_utils import Pix2PixHDModel, Pix2PixHDLoss, Pix2PixHDTrainer
    from fastai.vision import Learner, partial, optim
    from .._data_utils.pix2pix_data import show_results, predict
    from .._utils.common import _get_emd_path, get_multispectral_data_params_from_emd
    from ._pix2pix_utils import compute_fid_metric, compute_metrics
    from pathlib import Path
    import torch

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class Pix2PixHD(ArcGISModel):

    """
    Creates a model object which generates fake images of type B from type A.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    n_gen_filters           Optional int. Number of gen filters in first conv layer.
                            Default: 64
    ---------------------   -------------------------------------------
    gen_network             Optional string (global/local). Selects model to use for generator.
                            Use global if gpu memory is less.
                            Default: "local"
    ---------------------   -------------------------------------------
    n_downsample_global     Optional int. Number of downsampling layers in gen_network
                            Default: 4
    ---------------------   -------------------------------------------
    n_blocks_global         Optional int. Number of residual blocks in the global
                            generator network.
                            Default: 9
    ---------------------   -------------------------------------------
    n_local_enhancers       Optional int. Number of local enhancers to use.
                            Default: 1
    ---------------------   -------------------------------------------
    n_blocks_local          Optional int. number of residual blocks in the local
                            enhancer network.
                            Default: 3
    ---------------------   -------------------------------------------
    norm                    Optional string. instance normalization or batch normalization
                            Default: "instance"
    ---------------------   -------------------------------------------
    lsgan                   Optional bool. Use least square GAN, if True,
                            use vanilla GAN.
                            Default: True
    ---------------------   -------------------------------------------
    n_dscr_filters          Optional int. number of discriminator filters in first conv layer.
                            Default: 64
    ---------------------   -------------------------------------------
    n_layers_dscr           Optional int. only used if which_model_net_dscr==n_layers.
                            Default: 3
    ---------------------   -------------------------------------------
    n_dscr                  Optional int. number of discriminators to use.
                            Default: 2
    ---------------------   -------------------------------------------
    feat_loss               Optional bool. if 'True', use discriminator
                            feature matching loss.
                            Default: True
    ---------------------   -------------------------------------------
    vgg_loss                Optional bool. if 'True', use VGG feature matching loss.
                            Default: True (supported for 3 band imagery only).
    ---------------------   -------------------------------------------
    lambda_feat             Optional int. weight for feature matching loss.
                            Default: 10
    ---------------------   -------------------------------------------
    lambda_l1               Optional int. weight for feature matching loss.
                            Default: 100 (not supported for 3 band imagery)
    =====================   ===========================================

    :return: :class:`~arcgis.learn.Pix2PixHD` Object
    """

    def __init__(self, data, pretrained_path=None, *args, **kwargs):
        super().__init__(data, pretrained_path=pretrained_path, **kwargs)
        self._check_dataset_support(data)
        # input_nc=3, output_nc=3,
        self.kwargs = kwargs
        vgg_loss = kwargs.get("vgg_loss", True)
        lambda_feat = kwargs.get("lambda_feat", 10.0)
        l1_loss = kwargs.get("l1_loss", False)
        lambda_l1 = kwargs.get("lambda_l1", 100)

        self.input_nc, self.output_nc = 3, 3
        label_nc = self._data.label_nc
        if self._data._is_multispectral:
            self.input_nc, self.output_nc = self._data.n_channel, self._data.n_channel
            vgg_loss = False
            l1_loss = True
        elif self._data.label_nc:
            self.input_nc = label_nc
        if self._device.type == "cuda":
            gpu_ids = [torch.cuda.current_device()]
        else:
            gpu_ids = []
        pix2pix_hd = Pix2PixHDModel(
            label_nc, self.input_nc, self.output_nc, gpu_ids, **kwargs
        )

        self.learn = Learner(
            data,
            pix2pix_hd,
            loss_func=Pix2PixHDLoss(
                pix2pix_hd, vgg_loss, lambda_feat, l1_loss, lambda_l1, gpu_ids
            ),
            callback_fns=[Pix2PixHDTrainer],
            opt_func=partial(optim.Adam, betas=(0.5, 0.99)),
        )

        self.learn.model = self.learn.model.to(self._device)
        self._slice_lr = False
        if pretrained_path is not None:
            self.load(pretrained_path)
        self._code = image_translation_prf

        def __str__(self):
            return self.__repr__()

        def __repr__(self):
            return "<%s>" % (type(self).__name__)

    def show_results(self, rows=2, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        rgb_bands               Optional list of integers (band numbers)
                                to be considered for rgb visualization.
        =====================   ===========================================

        """
        show_results(self, rows, **kwargs)

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.Pix2PixHD` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.Pix2PixHD` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        resize_to = emd.get("resize_to")
        chip_size = emd["ImageHeight"]
        norm_stats = emd.get("norm_stats")
        kwargs = emd.get("Kwargs", {})
        _ = [kwargs.pop(key) for key in ["backbone", "backend"] if key in kwargs.keys()]
        if emd.get("ArcGISLearnVersion") < "2.0.1":
            if "gen_network" not in kwargs:
                kwargs["gen_network"] = "global"

        if data is None:
            data = _EmptyData(
                path=emd_path.parent, loss_func=None, c=2, chip_size=resize_to
            )
            if emd.get("IsMultispectral", False):
                data = get_multispectral_data_params_from_emd(data, emd)

                normalization_stats_b = dict(emd.get("NormalizationStats_b"))
                for _stat in normalization_stats_b:
                    if normalization_stats_b[_stat] is not None:
                        normalization_stats_b[_stat] = torch.tensor(
                            normalization_stats_b[_stat]
                        )
                    setattr(data, ("_" + _stat), normalization_stats_b[_stat])

            data._is_multispectral = emd.get("IsMultispectral", False)
            data.n_channel = emd.get("n_intput_channel", None)
            if data.n_channel == None:
                data.n_channel = emd.get("n_channel", None)
            data.label_nc = emd.get("label_nc", None)
            data.output_nc = emd.get("output_nc", None)
            data.mask_map = emd.get("mask_map", None)
            data.emd_path = emd_path
            data.emd = emd
            data._is_empty = True
            data.resize_to = chip_size
            data.norm_stats = norm_stats
            data.imagery_type = emd.get("ImageryType")

        return cls(data, **model_params, pretrained_path=str(model_file), **kwargs)

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_pix2pix_hd"
        _emd_template["InferenceFunction"] = "ArcGISImageTranslation.py"
        _emd_template["ModelType"] = "Pix2PixHD"
        _emd_template["n_intput_channel"] = self.output_nc
        _emd_template["label_nc"] = self._data.label_nc
        if self._data.label_nc != 0:
            _emd_template["mask_map"] = self._data.mask_map.tolist()
        _emd_template["Kwargs"] = self.kwargs
        # _emd_template["input_nc"] = self.input_nc

        norm_stats = []
        for k in self._data.norm_stats:
            norm_stats.append(k)
        _emd_template["norm_stats"] = list(norm_stats)
        # _emd_template["SupportsVariableTileSize"] = True
        # if self._data._is_multispectral:
        _emd_template["NormalizationStats_b"] = {
            "band_min_values": self._data._band_min_values_b,
            "band_max_values": self._data._band_max_values_b,
            "band_mean_values": self._data._band_mean_values_b,
            "band_std_values": self._data._band_std_values_b,
            "scaled_min_values": self._data._scaled_min_values_b,
            "scaled_max_values": self._data._scaled_max_values_b,
            "scaled_mean_values": self._data._scaled_mean_values_b,
            "scaled_std_values": self._data._scaled_std_values_b,
        }
        for _stat in _emd_template["NormalizationStats_b"]:
            if _emd_template["NormalizationStats_b"][_stat] is not None:
                _emd_template["NormalizationStats_b"][_stat] = _emd_template[
                    "NormalizationStats_b"
                ][_stat].tolist()
        _emd_template["n_channel"] = len(
            _emd_template["NormalizationStats_b"]["band_min_values"]
        )
        return _emd_template

    @property
    def _model_metrics(self):
        return self.compute_metrics()

    def predict(self, path):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        img_path                Required path of an image.
        =====================   ===========================================

        """
        return predict(self, path)

    def compute_metrics(self, accuracy=True, show_progress=True):
        """
        Computes Peak Signal-to-Noise Ratio (PSNR) and
        Structural Similarity Index Measure (SSIM) on validation set.
        Additionally, computes Frechet Inception Distance (FID) for
        RGB imagery only.

        """
        psnr, ssim = compute_metrics(self, self._data.valid_dl, show_progress)
        if self._data._imagery_type_b == "RGB" and self._data.n_channel == 3:
            fid = compute_fid_metric(self, self._data)
            return {
                "PSNR": "{0:1.4e}".format(psnr),
                "SSIM": "{0:1.4e}".format(ssim),
                "FID": "{0:1.4e}".format(fid),
            }
        else:
            fid = None
            return {"PSNR": "{0:1.4e}".format(psnr), "SSIM": "{0:1.4e}".format(ssim)}

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return Pix2PixHD._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Pix2Pix", "Export_Tiles"]



from __future__ import annotations
from arcgis.gis import Item, Group, GIS
from arcgis._impl.common._clone import (
    CloneNode,
    _ItemDefinition,
    _TextItemDefinition,
)


class BaseCloneDefinition(CloneNode):
    """
    The base cloning module that allows users to extend the cloning API to
    meet there cloning workflows.
    """

    def clone(self) -> Item:
        """
        Override the clone operation in order performs the cloning logic
        """
        raise NotImplementedError("clone is not implemented")


class BaseCloneItemDefinition(_ItemDefinition):
    """
    Represents the definition of an item within ArcGIS Online or Portal.
    """

    def clone(self) -> Item:
        """
        Override the clone operation in order performs the cloning logic
        """
        raise NotImplementedError("clone is not implemented")


class BaseCloneTextItemDefinition(_TextItemDefinition):
    """
    Represents the definition of a text based item within ArcGIS Online or Portal.
    """

    def clone(self) -> Item:
        """
        Override the clone operation in order performs the cloning logic
        """
        raise NotImplementedError("clone is not implemented")


class BaseCloneGroup:
    """
    The base cloning module that allows users to extend the cloning API to
    meet there cloning workflows.
    """

    _group_source: Group = None
    _include_items: bool = None
    _gis: GIS = None

    def __init__(
        self,
        *,
        gis: GIS | None = None,
    ) -> None:
        """initializer"""
        if gis is None:
            import arcgis

            gis = arcgis.env.active_gis
        self._gis: GIS = gis

    def clone(self, **kwags) -> Group:
        """
        Override the clone operation in order performs the cloning logic
        """
        raise NotImplementedError("clone is not implemented")



import os, random, sys, math
import json
import numpy as np
from pathlib import Path
from fastai.data_block import get_files as gf
import torch
from fastai.vision import (
    open_mask,
    get_transforms,
    plt,
    Image,
    ImageSegment,
    image2np,
)
from torch.utils.data import Dataset, DataLoader
from fastai.data_block import DataBunch
from fastai.vision.image import _resolve_tfms
import arcgis
from .._utils.common import (
    get_nbatches,
    get_top_padding,
    ArcGISMSImage,
    ArcGISImageList,
)
import types
from functools import partial
from .._data import _prepare_working_dir
from .._utils.cyclegan import image_extensions
from .._data import _tensor_scaler
from .._utils.superres import show_batch
import warnings

stats = [[0.5, 0.5, 0.5], [0.5, 0.5, 0.5]]


def get_mask_map(img_paths):
    img_list = []
    for img_path in img_paths:
        img = open_mask(img_path)
        img = img.data.numpy()
        img_list.append(np.unique(img))
    img_list = np.concatenate(img_list)
    img_list = np.sort(img_list)
    img_list = np.unique(img_list)
    return img_list


def rescale_mask(mask, map):
    mask = mask.data
    for i, j in enumerate(map):
        mask[mask == j] = i
    mask = mask.float()
    return ImageSegment(mask)


def get_device():
    if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
        device = torch.device("cuda")
    elif getattr(arcgis.env, "_processorType", "") == "CPU":
        device = torch.device("cpu")
    else:
        device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )

    return device


def is_old_format_pix2pix(path):
    """
    Function that returns 'True' if the manually created dir structure is being used.
    """
    return os.path.exists(path / "Images" / "train_a") and os.path.exists(
        path / "Images" / "train_b"
    )


def is_thematic_data(path):
    """
    Function that returns 'True' if thematic labels are present instead of rgb as source image
    """
    return os.path.exists(path / "labels") and os.path.exists(path / "images")


def pix2pix_paths(path):
    """
    Function to return the appropriate image dir paths in the provided dataset dir.
    """
    if is_old_format_pix2pix(path):
        return (path / "Images" / "train_a", path / "Images" / "train_b")
    elif is_thematic_data(path):  # only works with Pix2PixHD
        return (path / "labels", path / "images")
    else:
        return (path / "images", path / "images2")


def folder_check_pix2pix(path):
    """
    Function to check if the correct dir structure is provided.
    """
    img_folder1 = os.path.exists(path / "Images" / "train_a") or os.path.exists(
        path / "images"
    )
    img_folder2 = (
        os.path.exists(path / "Images" / "train_b")
        or os.path.exists(path / "images2")
        or os.path.exists(path / "labels")
    )
    if not all([img_folder1, img_folder2]):
        raise Exception(
            f"""You might be using an incorrect format to train your model. \nPlease ensure your training data has the following folder structure:
                ├─dataset folder name
                    ├─images
                    ├─images2   """
        )


def _tensor_scaler_tfm(tensor_batch, min_values, max_values, mode="minmax"):
    x = tensor_batch
    if x.shape[0] > min_values.shape[0]:
        res = x.shape[0] - min_values.shape[0]
        last_val = torch.tensor([min_values[min_values.shape[0] - 1]])
        for i in range(res):
            min_values = torch.tensor(np.concatenate((min_values, last_val), axis=0))
    if x.shape[0] > max_values.shape[0]:
        res = x.shape[0] - max_values.shape[0]
        last_val = torch.tensor([max_values[max_values.shape[0] - 1]])
        for i in range(res):
            max_values = torch.tensor(np.concatenate((max_values, last_val), axis=0))
    max_values = max_values.view(-1, 1, 1).to(x.device)
    min_values = min_values.view(-1, 1, 1).to(x.device)
    x = _tensor_scaler(x, min_values, max_values, mode, create_view=False)
    return x


def _norm_stats(path):
    emd_path = Path(os.path.abspath(path / "esri_model_definition.emd"))
    with open(emd_path) as f:
        emd_stats = json.load(f)

    domain_stats1 = emd_stats.get("AllTilesStats")
    domain_stats2 = emd_stats.get("AllTilesStats2", domain_stats1)

    data_stats = [domain_stats1, domain_stats2]
    batch_stats = [
        {
            "band_min_values": torch.tensor([i.get("Min") for i in stats]),
            "band_max_values": torch.tensor([i.get("Max") for i in stats]),
            "band_mean_values": torch.tensor([i.get("Mean") for i in stats]),
            "band_std_values": torch.tensor([i.get("StdDev") for i in stats]),
            "scaled_min_values": None,
            "scaled_max_values": None,
            "scaled_mean_values": None,
            "scaled_std_values": None,
        }
        for stats in data_stats
    ]
    return batch_stats[0], batch_stats[1]


def _batch_stats_json(
    path, img_list, norm_pct, stats_file_name="esri_normalization_stats.json"
):
    from .._data import _get_batch_stats

    if len(img_list) < 300:
        norm_pct = 1

    dummy_stats = {
        "batch_stats_for_norm_pct_0": {
            "band_min_values": None,
            "band_max_values": None,
            "band_mean_values": None,
            "band_std_values": None,
            "scaled_min_values": None,
            "scaled_max_values": None,
            "scaled_mean_values": None,
            "scaled_std_values": None,
        }
    }

    normstats_json_path = os.path.abspath(path / stats_file_name)

    if not os.path.exists(normstats_json_path):
        normstats = dummy_stats
        with open(normstats_json_path, "w", encoding="utf-8") as f:
            json.dump(normstats, f, ensure_ascii=False, indent=4)
    else:
        with open(normstats_json_path) as f:
            normstats = json.load(f)

    emd_path = Path(os.path.abspath(path / "esri_model_definition.emd"))
    with open(emd_path) as f:
        emd_stats = json.load(f)
    if "AllTilesStats2" not in emd_stats:
        emd_stats = None
    else:
        if stats_file_name == "esri_normalization_stats_a.json":
            domain_stats = emd_stats.get("AllTilesStats")
        else:
            domain_stats = emd_stats.get("AllTilesStats2")
        mini, maxi, mean, std = [], [], [], []
        for i in domain_stats:
            mini.append(i.get("Min"))
            maxi.append(i.get("Max"))
            mean.append(i.get("Mean"))
            std.append(i.get("StdDev"))

        data_stats = [mini, maxi, mean, std]
        data_stats_tensors = [
            torch.tensor(mini),
            torch.tensor(maxi),
            torch.tensor(mean),
            torch.tensor(std),
        ]

    norm_pct_search = f"batch_stats_for_norm_pct_{round(norm_pct*100)}"
    if norm_pct_search in normstats:
        batch_stats = normstats[norm_pct_search]
        if emd_stats is not None:
            for l, s in enumerate(batch_stats):
                if l == len(data_stats):
                    break
                batch_stats[s] = data_stats[l]

            for s in batch_stats:
                if batch_stats[s] is not None:
                    batch_stats[s] = torch.tensor(batch_stats[s])
        else:
            for s in batch_stats:
                if batch_stats[s] is not None:
                    batch_stats[s] = torch.tensor(batch_stats[s])

    else:
        if emd_stats is not None:
            batch_stats = {
                "band_min_values": None,
                "band_max_values": None,
                "band_mean_values": None,
                "band_std_values": None,
                "scaled_min_values": None,
                "scaled_max_values": None,
                "scaled_mean_values": None,
                "scaled_std_values": None,
            }
            for l, s in enumerate(batch_stats):
                if l == len(data_stats):
                    break
                batch_stats[s] = data_stats[l]

            for s in batch_stats:
                if batch_stats[s] is not None:
                    batch_stats[s] = torch.Tensor(batch_stats[s])
        else:
            batch_stats = _get_batch_stats(img_list, norm_pct)
            for s in batch_stats:
                if batch_stats[s] is not None:
                    batch_stats[s] = torch.Tensor(batch_stats[s])
        normstats[norm_pct_search] = dict(batch_stats)
        for s in normstats[norm_pct_search]:
            if normstats[norm_pct_search][s] is not None:
                normstats[norm_pct_search][s] = normstats[norm_pct_search][s].tolist()
        with open(normstats_json_path, "w", encoding="utf-8") as f:
            json.dump(normstats, f, ensure_ascii=False, indent=4)

    return batch_stats


def multispectral_additions(data):
    # batch_stats_a = _batch_stats_json(data.path, data.x_a, norm_pct, stats_file_name="esri_normalization_stats_a.json")
    # batch_stats_b = _batch_stats_json(data.path, data.x_b, norm_pct, stats_file_name="esri_normalization_stats_b.json")

    data._band_min_values = data.batch_stats_a["band_min_values"]
    data._band_max_values = data.batch_stats_a["band_max_values"]
    data._band_mean_values = data.batch_stats_a["band_mean_values"]
    data._band_std_values = data.batch_stats_a["band_std_values"]
    data._scaled_min_values = data.batch_stats_a["scaled_min_values"]
    data._scaled_max_values = data.batch_stats_a["scaled_max_values"]
    data._scaled_mean_values = data.batch_stats_a["scaled_mean_values"]
    data._scaled_std_values = data.batch_stats_a["scaled_std_values"]

    data._band_min_values_b = data.batch_stats_b["band_min_values"]
    data._band_max_values_b = data.batch_stats_b["band_max_values"]
    data._band_mean_values_b = data.batch_stats_b["band_mean_values"]
    data._band_std_values_b = data.batch_stats_b["band_std_values"]
    data._scaled_min_values_b = data.batch_stats_b["scaled_min_values"]
    data._scaled_max_values_b = data.batch_stats_b["scaled_max_values"]
    data._scaled_mean_values_b = data.batch_stats_b["scaled_mean_values"]
    data._scaled_std_values_b = data.batch_stats_b["scaled_std_values"]


def get_files(*args, **kwargs):
    return sorted(gf(*args, **kwargs))


def apply_tfms(images, other_tfms, resize_to):
    image_A, image_B = images
    # To fix when the disk image size is smaller than
    # crop size, otherwise crop_tfm behaves strange    if min(image_B.shape[1:]) / size < 1:
    #         image_A = image_A.resize(size)
    #         image_B = image_B.resize(size
    # )
    # image_inst = image_inst.resize(size)
    image_A = image_A.resize((3, resize_to, resize_to))
    image_B = image_B.resize((3, resize_to, resize_to))

    image_A = image_A.apply_tfms(other_tfms, do_resolve=False)

    image_B = image_B.apply_tfms(other_tfms, do_resolve=False)
    return image_A, image_B


def concat_bands(image_A, image_B):
    if image_A.shape[0] < image_B.shape[0]:
        cont = []
        last_tile = np.expand_dims(image_A.data[image_A.shape[0] - 1, :, :], 0)
        res = abs(image_B.shape[0] - image_A.shape[0])
        for i in range(res):
            image_A = ArcGISMSImage(
                torch.tensor(np.concatenate((image_A.data, last_tile), axis=0))
            )
    if image_B.shape[0] < image_A.shape[0]:
        cont = []
        last_tile = np.expand_dims(image_B.data[image_B.shape[0] - 1, :, :], 0)
        res = abs(image_A.shape[0] - image_B.shape[0])
        for i in range(res):
            image_B = ArcGISMSImage(
                torch.tensor(np.concatenate((image_B.data, last_tile), axis=0))
            )

    return image_A, image_B


def _get_transforms(transforms, model=None):
    if transforms is None:
        if model:
            transforms = get_transforms(
                do_flip=True,
                flip_vert=True,
                max_rotate=90.0,
                max_zoom=0,
                max_lighting=None,
                max_warp=None,
                p_affine=0.75,
                xtra_tfms=None,
            )
        else:
            transforms = get_transforms(flip_vert=True, max_lighting=0.3, max_warp=0.0)
    elif transforms is False:
        transforms = ([], [])

    return transforms


# images_A = labels
# images_B = rgb_imgs
# images_inst = instance_labels


class Pix2PixHDDataset(Dataset):
    def __init__(
        self,
        path,
        image_list_A,
        image_list_B,
        resize_to,
        transforms=None,
        norm_stats=None,
        flip_vert=True,
        split=None,
        label_nc=False,
        _is_multispectral=False,
        imagery_type="RGB",
        batch_stats_a=None,
        batch_stats_b=None,
        rgb_bands=[0, 1, 2],
    ):
        self.path = Path(path)
        self.image_list_A = image_list_A
        self.image_list_B = image_list_B
        # self.image_list_inst = image_list_inst

        self.train_tfms, self.val_tfms = _get_transforms(transforms)
        self.split = split
        self.norm_stats = norm_stats
        self.label_nc = label_nc
        self.x_a = ArcGISImageList(image_list_A)
        self.x_b = ArcGISImageList(image_list_B)
        if resize_to is None:
            resize_to = self.x_a[0].data.shape[-1]
        if label_nc:  # fix for https://github.com/NVIDIA/pix2pixHD/issues/100
            if resize_to % 32 != 0:
                resize_to = resize_to - resize_to % 32
        self.resize_to = resize_to
        self.y = 0

        # MS
        self._is_multispectral = _is_multispectral
        self.n_channel = (
            self.x_a[0].shape[0]
            if self.x_a[0].shape[0] > self.x_b[0].shape[0]
            else self.x_b[0].shape[0]
        )
        if self._is_multispectral:
            self.norm_stats = [[0.5] * self.n_channel, [0.5] * self.n_channel]
        self.batch_stats_a = batch_stats_a
        self.batch_stats_b = batch_stats_b
        self.imagery_type = imagery_type
        self.rgb_bands = rgb_bands
        if label_nc:
            self.mask_map = get_mask_map(image_list_A)

    def __len__(self):
        return len(self.image_list_B)

    def __getitem__(self, idx):
        if os.path.isdir(self.path / "labels"):
            self.image_list_B = [
                Path(
                    os.path.join(
                        os.path.split(i)[0].replace("images2", "labels"),
                        os.path.split(i)[1],
                    )
                )
                for i in self.image_list_B
            ]
        if self._is_multispectral:
            # to add mask loading
            image_A = ArcGISMSImage.open(
                self.image_list_A[idx], imagery_type=self.imagery_type
            )
            image_B = ArcGISMSImage.open(
                self.image_list_B[idx], imagery_type=self.imagery_type
            )

            image_A, image_B = concat_bands(image_A, image_B)

            image_A = _tensor_scaler_tfm(
                image_A.data,
                min_values=self.batch_stats_a["band_min_values"],
                max_values=self.batch_stats_a["band_max_values"],
                mode="minmax",
            )
            image_B = _tensor_scaler_tfm(
                image_B.data,
                min_values=self.batch_stats_b["band_min_values"],
                max_values=self.batch_stats_b["band_max_values"],
                mode="minmax",
            )

            image_A = ArcGISMSImage(image_A)
            image_B = ArcGISMSImage(image_B)

        else:
            min_vals_a = self.batch_stats_a["band_min_values"].tolist()
            max_vals_a = self.batch_stats_a["band_max_values"].tolist()
            min_vals_b = self.batch_stats_b["band_min_values"].tolist()
            max_vals_b = self.batch_stats_b["band_max_values"].tolist()
            rgb_mins_a, rgb_maxs_a = (
                [float(i) for i in min_vals_a],
                [float(i) for i in max_vals_a],
            )
            rgb_mins_b, rgb_maxs_b = (
                [float(i) for i in min_vals_b],
                [float(i) for i in max_vals_b],
            )
            if self.label_nc == False:
                image_A = ArcGISMSImage.open(
                    self.image_list_A[idx],
                    imagery_type=self.imagery_type,
                    div=(rgb_mins_a, rgb_maxs_a),
                )
            else:
                image_A = ImageSegment(ArcGISMSImage.open(self.image_list_A[idx]).data)
                image_A = rescale_mask(image_A, self.mask_map)

            image_B = ArcGISMSImage.open(
                self.image_list_B[idx],
                imagery_type=self.imagery_type,
                div=(rgb_mins_b, rgb_maxs_b),
            )

        _resolve_tfms(self.train_tfms)
        _resolve_tfms(self.val_tfms)

        images = (image_A, image_B)
        if self.split == "train":
            images = apply_tfms(images, self.train_tfms, self.resize_to)
        else:
            images = apply_tfms(images, self.val_tfms, self.resize_to)
        image_A, image_B = images

        if self.label_nc == False:
            self.image_A = image_A
            image_A = normalize(image_A.px, *self.norm_stats)
            # image_A = -1+2*image_A.data
        else:
            self.image_A = image_A
            image_A = image_A.data

        self.image_B = image_B
        image_B = normalize(image_B.px, *self.norm_stats)
        return (image_A, image_B), self.y

    def __repr__(self):
        item = self.__getitem__(0)
        return f"{self.__class__.__name__}{(item[0][0].shape, item[0][1].shape)}, items_A:{len(self.image_list_A)}, items_B:{len(self.image_list_B)}"

    def show(self, idx, axes=None, rgb_bands=None):
        if rgb_bands is None:
            rgb_bands = self.rgb_bands
        self[idx]
        if axes is None:
            _, axes = plt.subplots(1, 2, figsize=(15, 7))

        if self.label_nc == True:
            self.image_A.show(axes[0])
        else:
            self.image_A.show(axes[0], rgb_bands=rgb_bands)
        self.image_B.show(axes[1], rgb_bands=rgb_bands)


class SR3Dataset(Pix2PixHDDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_multispec = self._is_multispectral
        self._n_channel = self.x_a[0].shape[0]
        self.resize_to = self.x_b[0].data.shape[-1]
        self._image_stats, self._image_stats2 = None, None
        self.train_tfms, self.val_tfms = _get_transforms(transforms=None, model="SR3")

    def __getitem__(self, idx):
        min_vals_a = self.batch_stats_a["band_min_values"]
        max_vals_a = self.batch_stats_a["band_max_values"]
        min_vals_b = self.batch_stats_b["band_min_values"]
        max_vals_b = self.batch_stats_b["band_max_values"]
        if self._is_multispectral:
            # to add mask loading
            image_A = ArcGISMSImage.open(
                self.image_list_A[idx], imagery_type=self.imagery_type
            )
            image_B = ArcGISMSImage.open(
                self.image_list_B[idx], imagery_type=self.imagery_type
            )

            image_A, image_B = concat_bands(image_A, image_B)

            image_A = _tensor_scaler_tfm(
                image_A.data,
                min_values=min_vals_a,
                max_values=max_vals_a,
                mode="minmax",
            )
            image_B = _tensor_scaler_tfm(
                image_B.data,
                min_values=min_vals_b,
                max_values=max_vals_b,
                mode="minmax",
            )

            image_A = ArcGISMSImage(image_A)
            image_B = ArcGISMSImage(image_B)

        else:
            image_A = ArcGISMSImage.open(
                self.image_list_A[idx],
                imagery_type=self.imagery_type,
                div=255,
            )
            image_B = ArcGISMSImage.open(
                self.image_list_B[idx],
                imagery_type=self.imagery_type,
                div=255,
            )

        _resolve_tfms(self.train_tfms)
        _resolve_tfms(self.val_tfms)

        images = (image_A, image_B)
        if self.split == "train":
            images = apply_tfms(images, self.train_tfms, self.resize_to)
        else:
            images = apply_tfms(images, self.val_tfms, self.resize_to)
        image_A, image_B = images

        self.image_A = image_A
        image_A = -1 + 2 * image_A.px  # normalize(image_A.px, *self.norm_stats)
        image_A = image_A.data

        self.image_B = image_B
        image_B = -1 + 2 * image_B.px  # normalize(image_B.px, *self.norm_stats)
        image_B = image_B.data
        return (image_A, image_B), image_B

    def __repr__(self):
        item = self.__getitem__(0)
        return f"{self.__class__.__name__}{(item[0][0].shape, item[0][1].shape)}, items_A:{len(self.image_list_A)}, items_B:{len(self.image_list_B)}"


def create_train_val_sets(
    path,
    val_split_pct,
    transforms,
    resize_to,
    norm_stats,
    flip_vert,
    label_nc=False,
    _is_multispectral=False,
    rgb_bands=None,
    norm_pct=1,
    model_type="Pix2Pix",
    **kwargs,
):
    path_lr = kwargs.get("path_lr", None)
    path_hr = kwargs.get("path_hr", None)

    imagery_type = kwargs.get("imagery_type", "None")
    path = Path(path)
    if model_type == "Pix2Pix":
        path_A, path_B = pix2pix_paths(path)
    else:
        path_A, path_B = path_lr, path_hr

    images_A = get_files(path_A, extensions=image_extensions, recurse=True)

    images_B = get_files(path_B, extensions=image_extensions, recurse=True)

    batch_stats_a, batch_stats_b = None, None

    if model_type == "Pix2Pix":
        batch_stats_a = _batch_stats_json(
            path,
            ArcGISImageList(images_A),
            norm_pct,
            stats_file_name="esri_normalization_stats_a.json",
        )
        batch_stats_b = _batch_stats_json(
            path,
            ArcGISImageList(images_B),
            norm_pct,
            stats_file_name="esri_normalization_stats_b.json",
        )
    else:
        batch_stats_a, batch_stats_b = _norm_stats(path)

    total_num_images = len(images_B)
    val_num_images = int(total_num_images * val_split_pct)
    zipped = list(zip(images_A, images_B))

    random.shuffle(zipped)

    images_A, images_B = zip(*zipped)

    train_images_A, train_images_B = (
        images_A[val_num_images:],
        images_B[val_num_images:],
    )

    val_images_A, val_images_B = (images_A[:val_num_images], images_B[:val_num_images])

    if model_type == "Pix2Pix":
        datasetclass = Pix2PixHDDataset
    else:
        datasetclass = SR3Dataset

    train_dataset = datasetclass(
        path,
        train_images_A,
        train_images_B,
        transforms=transforms,
        resize_to=resize_to,
        norm_stats=norm_stats,
        flip_vert=flip_vert,
        split="train",
        label_nc=label_nc,
        _is_multispectral=_is_multispectral,
        imagery_type=imagery_type,
        batch_stats_a=batch_stats_a,
        batch_stats_b=batch_stats_b,
        rgb_bands=rgb_bands,
    )
    val_dataset = datasetclass(
        path,
        val_images_A,
        val_images_B,
        transforms=transforms,
        resize_to=resize_to,
        norm_stats=norm_stats,
        flip_vert=flip_vert,
        split="val",
        label_nc=label_nc,
        _is_multispectral=_is_multispectral,
        imagery_type=imagery_type,
        batch_stats_a=batch_stats_a,
        batch_stats_b=batch_stats_b,
        rgb_bands=rgb_bands,
    )

    return train_dataset, val_dataset


def create_dataloaders(datasets, batch_size, dataloader_kwargs):
    return [DataLoader(d, batch_size, **dataloader_kwargs) for d in datasets]


def prepare_pix2pix_data(
    path,
    batch_size,
    val_split_pct,
    transforms,
    resize_to,
    norm_pct,
    _is_multispectral,
    working_dir,
    seed,
    dataset_type,
    **kwargs,
):
    norm_stats = [[0.5, 0.5, 0.5], [0.5, 0.5, 0.5]]  # kwargs.get('norm_stats', stats)
    flip_vert = kwargs.get("imagery_type", "satellite") != "oriented"
    split = kwargs.get("split", "random")
    if dataset_type == "Pix2Pix":
        label_nc = is_thematic_data(path) or kwargs.get("label_nc", False)
    else:
        label_nc = False
    if label_nc:
        _is_multispectral = False
    imagery_type = kwargs.get("imagery_type")
    rgb_bands = kwargs.get("rgb_bands")
    datasets = create_train_val_sets(
        path,
        val_split_pct,
        transforms=transforms,
        resize_to=resize_to,
        norm_stats=norm_stats,
        flip_vert=flip_vert,
        label_nc=label_nc,
        _is_multispectral=_is_multispectral,
        rgb_bands=rgb_bands,
        norm_pct=norm_pct,
        model_type=dataset_type,
        **kwargs,
    )

    databunch_kwargs = (
        {"num_workers": 0}
        if sys.platform == "win32"
        else {"num_workers": os.cpu_count() - 4}
    )

    train_dl, valid_dl = create_dataloaders(datasets, batch_size, databunch_kwargs)
    device = get_device()
    data = DataBunch(train_dl, valid_dl, device=device)
    data.path = data.train_ds.path
    data._do_normalize = None
    data._imagery_type = imagery_type
    data._norm_pct = norm_pct
    # data.n_channel = data.train_ds[0][0][0].shape[0]
    multispectral_additions(data)
    if label_nc:
        data.label_nc = len(data.mask_map) + 1
    data.chip_size = data.resize_to
    if working_dir is not None:
        data.path = Path(os.path.abspath(working_dir))
    data._temp_folder = _prepare_working_dir(data.path)
    data.show_batch = types.MethodType(show_batch, data)
    data._dataset_type = dataset_type  # "Pix2Pix"
    data._downsampling_factor = kwargs.get("downsample_factor", None)
    data.val_split_pct = val_split_pct
    data.resize_to = resize_to
    data.seed = seed
    data._extract_bands = None

    return data


def show_batch(self, rows=4, **kwargs):
    """
    This function randomly picks a few training chips and visualizes them.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional int. Number of rows of results
                            to be displayed.
    =====================   ===========================================
    """
    rgb_bands = kwargs.get("rgb_bands", None)
    fig, axes = plt.subplots(nrows=rows, ncols=2, squeeze=False, figsize=(20, rows * 5))
    top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
    plt.subplots_adjust(top=top)
    # fig.suptitle("Input / Ground Truth", fontsize=16)
    axes[0, 0].title.set_text("Input")
    axes[0, 1].title.set_text("Target")
    img_idxs = [random.randint(0, len(self.train_ds) - 1) for k in range(rows)]
    for idx, im_idx in enumerate(img_idxs):
        self.train_ds.show(im_idx, axes[idx], rgb_bands)


def to_device(z, device):
    return torch.tensor(z).to(device)


def normalize(x, mean, std):
    "Normalize `x` with `mean` and `std`."
    to_tensor = partial(to_device, device=x.device)

    if type(mean[0]) is not torch.Tensor:
        mean = to_tensor(mean)
        std = to_tensor(std)
    return (x - mean[..., None, None]) / std[..., None, None]


def denormalize(x, mean, std, do_x=True):
    "Denormalize `x` with `mean` and `std`."
    to_tensor = partial(to_device, device=x.device)
    if type(mean[0]) is not torch.Tensor:
        mean = to_tensor(mean)
        std = to_tensor(std)
    return (
        x.cpu().float() * std[..., None, None] + mean[..., None, None]
        if do_x
        else x.cpu()
    )


def display_row(axes, display, rgb_bands=None):
    if rgb_bands is None:
        rgb_bands = [0, 1, 2]
    for i, ax in enumerate(axes):
        if type(display[i]) is ArcGISMSImage:
            display[i].show(ax, rgb_bands)
        else:
            ax.imshow(display[i])
        ax.axis("off")


def post_process(dists):
    if type(dists) is np.ndarray:
        dists = torch.from_numpy(dists).permute(2, 0, 1)
    return (dists > 1.0).long()


def show_results(self, rows, **kwargs):
    self.learn.model.eval()
    x_batch, y_batch = get_nbatches(
        self._data.valid_dl, math.ceil(rows / self._data.batch_size)
    )
    x_A, x_B = [x[0] for x in x_batch], [x[1] for x in x_batch]

    x_A = torch.cat(x_A)
    x_B = torch.cat(x_B)

    top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
    activ = []

    for i in range(0, x_B.shape[0], self._data.batch_size):
        with torch.no_grad():
            preds = self.learn.model(
                x_A[i : i + self._data.batch_size].detach(),
                x_B[i : i + self._data.batch_size].detach(),
            )
        activ.append(preds[0])

    activations = torch.cat(activ)
    if self._data.label_nc == 0:
        x_A = denormalize(x_A.cpu(), *self._data.norm_stats)
    x_B = denormalize(x_B.cpu(), *self._data.norm_stats)
    activations = denormalize(activations.cpu(), *self._data.norm_stats)
    rows = min(rows, x_A.shape[0])

    fig, axs = plt.subplots(
        nrows=rows, ncols=3, figsize=(4 * 5, rows * 5), squeeze=False
    )
    plt.subplots_adjust(top=top)
    axs[0, 0].title.set_text("Input")
    axs[0, 1].title.set_text("Target")
    axs[0, 2].title.set_text("Prediction")
    for r in range(rows):
        if self._data._is_multispectral:
            display_row(
                axs[r],
                (
                    ArcGISMSImage(x_A[r]),
                    ArcGISMSImage(x_B[r]),
                    ArcGISMSImage(activations[r]),
                ),
                kwargs.get("rgb_bands", None),
            )
        else:
            display_row(
                axs[r],
                (image2np(x_A[r]), image2np(x_B[r]), image2np(activations[r])),
            )


def predict(self, img_path):
    img_path = Path(img_path)
    if self._data.label_nc == 0:
        raw_img = ArcGISMSImage.open(
            img_path, imagery_type=self._data.imagery_type, div=255
        )
    else:
        raw_img = ImageSegment(ArcGISMSImage.open(img_path).data)
        raw_img = rescale_mask(raw_img, self._data.mask_map)
    raw_img = raw_img.resize(self._data.chip_size)

    if self._data._is_multispectral:
        n_band = self._data.n_channel
        if n_band > raw_img.shape[0]:
            cont = []
            last_tile = np.expand_dims(raw_img.data[raw_img.shape[0] - 1, :, :], 0)
            res = abs(n_band - raw_img.shape[0])
            for i in range(res):
                raw_img = Image(
                    torch.tensor(np.concatenate((raw_img.data, last_tile), axis=0))
                )

    if self._data.label_nc == 0:
        raw_img_tensor = normalize(raw_img.px, *self._data.norm_stats)
    else:
        raw_img_tensor = raw_img.px
    raw_img_tensor = raw_img_tensor[None].to(self._device)
    self.learn.model.eval()
    with torch.no_grad():
        prediction = (
            self.learn.model(raw_img_tensor, raw_img_tensor)[0].detach()[0].cpu()
        )

    pred_denorm = denormalize(prediction, *self._data.norm_stats)
    pred_denorm = ArcGISMSImage(pred_denorm)
    pred_denorm = pred_denorm.show()
    return pred_denorm


def rgb_or_ms(im_path):
    """
    Function that returns the imagery type (RGB or ms) of an image.
    """
    try:
        from osgeo import gdal

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ds = gdal.Open(im_path)
        if ds.RasterCount != 3 or ds.GetRasterBand(1).DataType != gdal.GDT_Byte:
            return "ms"
        else:
            return "RGB"
    except:
        return None
